# a framework for inference on multiple choice tasks.
import copy
import logging
import os

import torch
from torch.utils.data import DataLoader

from mm_poe.methods.utils.data import (
    upload_to_huggingface_hub,
    preprocess_function_seq2seq,
    preprocess_function_causal,
    preprocess_function_causal_channel,
    preprocess_function_seq2seq_channel,
    create_synonym_dataset,
    create_n_shot_splits,
)
from mm_poe.methods.utils.methods import (
    compute_conditional_score_seq2seq,
    compute_conditional_score_causal,
    inference_language_modeling,
    inference_calibration,
    inference_generate_synonyms,
    generate_synonyms,
)
from mm_poe.methods.utils.utils import (
    load_data,
    load_model,
    parse_args,
    set_seed,
    write_to_csv,
)

logger = logging.getLogger(__name__)


def main():
    # import pdb; pdb.set_trace()

    # step 1: argument parser, and logger
    args = parse_args()
    if args.multiple_choice_prompt is not None:
        args.method = "multiple_choice_prompt"
    elif args.amateur_checkpoint is not None:
        args.method = "contrastive_decoding"
    elif args.calibration_prompt is not None:
        args.method = "calibration"
    elif args.do_channel is True:
        args.method = "channel"
    elif args.do_synonym is True:
        args.method = "generate_synonyms"
    else:
        args.method = "language_modeling"

    # print(args)
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
    )
    logger.setLevel(logging.INFO)

    # step 2: set random seed to ensure reproducibility.
    logger.info(f"Set random seed to {args.seed}.")
    set_seed(args.seed)

    # step 3: load model, tokenizer.
    # Then move to gpu, and set to evaluation mode.
    logger.info(f"Load {args.model_family} model: {args.checkpoint}.")
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # get model path: ../models/args.model_family/args.checkpoint
    model_path = os.path.join("../models", args.model_family, args.checkpoint)
    model, tokenizer = load_model(device, model_path, args)
    if args.model_family in ["GPT2", "Pythia", "OPT-IML", "Dolly"]:
        compute_func = compute_conditional_score_causal
        preprocess_func = preprocess_function_causal
        preprocess_func_channel = preprocess_function_causal_channel
    elif args.model_family in ["T5", "FLAN-T5"]:
        compute_func = compute_conditional_score_seq2seq
        preprocess_func = preprocess_function_seq2seq
        preprocess_func_channel = preprocess_function_seq2seq_channel
    else:
        raise NotImplementedError

    # step 4: load and preprocess data.
    args.datasets = args.datasets.split()
    logger.info(f"Load data: {args.datasets}.")

    # evaluate on each dataset
    for dataset in args.datasets:
        args.dataset = dataset
        ending_names, header_name, raw_dataset, n_shot_dataset = load_data(
            args
        )
        raw_dataset, n_shot_dataset, n_shot_demonstrations = (
            create_n_shot_splits(raw_dataset, n_shot_dataset, args)
        )

        logger.info(f"Preprocess data: {args.dataset}.")
        fn_kwargs = {
            "ending_names": ending_names,
            "header_name": header_name,
            "tokenizer": tokenizer,
        }
        num_of_options = len(ending_names)
        tokenized_dataset = raw_dataset.map(
            preprocess_func,
            fn_kwargs=fn_kwargs,
            batched=True,
            batch_size=args.batch_size,
        )
        eval_dataloader = DataLoader(
            tokenized_dataset, batch_size=args.batch_size, shuffle=False
        )

        # step 5: (evaluation) inference on data, and compute accuracy.
        logger.info(
            f"Start inference (method: {args.method}) on {args.dataset} "
            + f"using {args.model_family} model: {args.checkpoint}."
        )
        if args.method in ["language_modeling", "multiple_choice_prompt"]:
            _, lm_accuracy, avg_lm_accuracy, _ = inference_language_modeling(
                model,
                eval_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        elif args.method == "contrastive_decoding":
            logger.info(
                f"Load {args.model_family} "
                + f"amateur model: {args.amateur_checkpoint}."
            )
            # get model path: ../models/args.model_family/args.checkpoint
            amateur_model_path = os.path.join(
                "../models", args.model_family, args.amateur_checkpoint
            )
            amateur_model, _ = load_model(device, amateur_model_path, args)
            # we want to integrate contrastive decoding with other methods,
            # so we need separate output from each model.
            # compute log probs on each model
            exp_avg_log_probs, exp_lm_accuracy, exp_avg_lm_accuracy, _ = (
                inference_language_modeling(
                    model,
                    eval_dataloader,
                    device,
                    compute_func,
                    tokenizer.pad_token_id,
                )
            )
            ama_avg_log_probs, ama_lm_accuracy, ama_avg_lm_accuracy, _ = (
                inference_language_modeling(
                    amateur_model,
                    eval_dataloader,
                    device,
                    compute_func,
                    tokenizer.pad_token_id,
                )
            )
            # calculate difference, and may introduce extra parameters.
            avg_log_probs = exp_avg_log_probs - ama_avg_log_probs
            labels = raw_dataset["label"]
            # currently, I use average language modeling accuracy.
            # I will add language modeling and other methods shortly.
            lm_accuracy = (
                avg_log_probs.argmin(dim=-1) == labels
            ).sum().item() / len(labels)
            logger.info(f"Contrastive decoding accuracy: {lm_accuracy:.4f}.")
            args.amateur_accuracy = ama_avg_lm_accuracy
            args.expert_accuracy = exp_avg_lm_accuracy
        elif args.method == "calibration":
            fn_kwargs = {
                "ending_names": ending_names,
                "header_name": "uncond_premise",  # the difference is here
                "tokenizer": tokenizer,
            }
            tokenized_calibration_dataset = raw_dataset.map(
                preprocess_func,
                fn_kwargs=fn_kwargs,
                batched=True,
                batch_size=args.batch_size,
            )
            eval_calibration_dataloader = DataLoader(
                tokenized_calibration_dataset,
                batch_size=args.batch_size,
                shuffle=False,
            )
            _, lm_accuracy, avg_lm_accuracy, _ = inference_calibration(
                model,
                eval_dataloader,
                eval_calibration_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        elif args.method == "channel":
            # simple solution: swap first sentence
            # and second sentence in both preprocessing functions
            tokenized_channel_dataset = raw_dataset.map(
                preprocess_func_channel,
                fn_kwargs=fn_kwargs,
                batched=True,
                batch_size=args.batch_size,
            )
            eval_channel_dataloader = DataLoader(
                tokenized_channel_dataset,
                batch_size=args.batch_size,
                shuffle=False,
            )
            _, lm_accuracy, avg_lm_accuracy, _ = inference_language_modeling(
                model,
                eval_channel_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        elif args.method == "generate_synonyms":
            # 3 stpes: generate synonyms, then map datasets, then inference.
            logger.info(f"Generate synonyms for {args.dataset}.")
            synonyms_dict = generate_synonyms(
                args, model, tokenizer, tokenized_dataset
            )
            # call map add synonyms to raw_dataset
            logger.info("Add synonyms to raw dataset.")
            synonym_kwargs = {
                "args": args,
                "synonyms_dict": synonyms_dict,
            }
            synonyms_dataset = raw_dataset.map(
                create_synonym_dataset,
                fn_kwargs=synonym_kwargs,
                batched=True,
                batch_size=args.batch_size,
            )
            # map to tokenized_dataset
            logger.info("Tokenize synonym data.")
            synonyms_ending_names = [
                col
                for col in synonyms_dataset.column_names
                if col.startswith("hypothesis")
            ]
            fn_kwargs = {
                "ending_names": synonyms_ending_names,
                "header_name": header_name,
                "tokenizer": tokenizer,
            }
            tokenized_synonyms_dataset = synonyms_dataset.map(
                preprocess_func,
                fn_kwargs=fn_kwargs,
                batched=True,
                batch_size=args.batch_size,
            )
            eval_synonyms_dataloader = DataLoader(
                tokenized_synonyms_dataset,
                batch_size=args.batch_size,
                shuffle=False,
            )
            _, lm_accuracy, avg_lm_accuracy, _ = inference_generate_synonyms(
                model,
                eval_synonyms_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
                num_of_options,
                args.number_of_synonyms,
            )
        else:
            raise NotImplementedError

        # step 6: some postprocessing, including saving and displyaing output.
        save_path = os.path.join("../results", f"{args.method}.csv")
        logger.info(f"Save results to {save_path}.")
        write_to_csv(save_path, args, lm_accuracy)

        # does calibration needs copying as well?
        if args.method == "language_modeling":
            avg_args = copy.deepcopy(args)
            avg_args.method = "average_language_modeling"
            write_to_csv(save_path, avg_args, avg_lm_accuracy)

        # step 7: push data to HuggingFace Hub.
        if args.push_data_to_hub:
            logger.info(f"Push {args.dataset} to HuggingFace Hub.")
            upload_to_huggingface_hub(tokenized_dataset, args)

        # step 8: delete tokenized_dataset to save memory.
        # del tokenized_dataset


if __name__ == "__main__":
    main()



def find_expert_model(model_family):
    expert_checkpoint = None
    if model_family == "GPT2":
        pass
    elif model_family == "T5":
        pass
    elif model_family == "FLAN-T5":
        expert_checkpoint = "google/flan-t5-base"
    else:
        print(f"{model_family}: Not implemented.")
    return expert_checkpoint



import os
import sys
import random
import numpy as np
import torch
import pytest
from unittest import mock
from unittest.mock import MagicMock, patch, mock_open

# Import the functions from utils.py
from mm_poe.methods.utils.utils import (
    set_seed,
    parse_args,
    load_data,
    load_model,
    write_to_csv,
)


# Test for set_seed function
@pytest.mark.parametrize("seed", [0, 42, 1234])
def test_set_seed(seed):
    with mock.patch("torch.cuda.manual_seed_all") as mock_cuda_seed_all:
        # Call the function with the seed
        set_seed(seed)

        # Check if os environment variable is set correctly
        assert os.environ["PYTHONHASHSEED"] == str(seed)

        # Check if random seeds are set correctly
        random_value = random.randint(0, 100)
        np_value = np.random.randint(0, 100)
        torch_value = torch.randint(0, 100, (1,)).item()

        set_seed(seed)  # Reset seeds

        assert random.randint(0, 100) == random_value
        assert np.random.randint(0, 100) == np_value
        assert torch.randint(0, 100, (1,)).item() == torch_value

        # Check if CUDA seeds were set correctly (mocked)
        mock_cuda_seed_all.assert_called_with(seed)


# Tests for parse_args function
def test_parse_args_with_required_arguments():
    test_args = [
        "script_name",
        "--model_family",
        "GPT2",
        "--checkpoint",
        "gpt2-medium",
        "--datasets",
        "copa",
    ]
    with mock.patch.object(sys, "argv", test_args):
        args = parse_args()
        assert args.model_family == "GPT2"
        assert args.checkpoint == "gpt2-medium"
        assert args.datasets == "copa"


def test_parse_args_with_all_arguments():
    test_args = [
        "script_name",
        "--model_family",
        "GPT2",
        "--checkpoint",
        "gpt2-medium",
        "--datasets",
        "copa winogrande",
        "--seed",
        "42",
        "--amateur_checkpoint",
        "gpt2-small",
        "--expert_method",
        "language_modeling",
        "--amateur_method",
        "calibration",
        "--weighting_parameter",
        "-0.5",
        "--weighting_parameters",
        "0.1,0.2",
        "--num_random_search",
        "5",
        "--loading_precision",
        "FP16",
        "--sample",
        "100",
        "--batch_size",
        "16",
        "--n_shot",
        "5",
        "--multiple_choice_prompt",
        "Choose the best option:",
        "--calibration_prompt",
        "This is a calibration prompt.",
        "--do_channel",
        "--process_of_elimination_prompt",
        "Eliminate incorrect options:",
        "--scoring_method_for_process_of_elimination",
        "calibration",
        "--prompting_method_for_process_of_elimination",
        "multiple_choice_prompt",
        "--mask_strategy_for_process_of_elimination",
        "min_k",
        "--do_synonym",
        "--number_of_synonyms",
        "3",
        "--generate_synonyms_prompt",
        "Generate synonyms for option: option",
        "--push_data_to_hub",
        "--min_k",
        "2",
        "--mask_token",
        "[MASK]",
    ]
    with mock.patch.object(sys, "argv", test_args):
        args = parse_args()
        assert args.seed == 42
        assert args.amateur_checkpoint == "gpt2-small"
        assert args.expert_method == "language_modeling"
        assert args.amateur_method == "calibration"
        assert args.weighting_parameter == -0.5
        assert args.weighting_parameters == "0.1,0.2"
        assert args.num_random_search == 5
        assert args.loading_precision == "FP16"
        assert args.sample == 100
        assert args.batch_size == 16
        assert args.n_shot == 5
        assert args.multiple_choice_prompt == "Choose the best option:"
        assert args.calibration_prompt == "This is a calibration prompt."
        assert args.do_channel is True
        assert (
            args.process_of_elimination_prompt
            == "Eliminate incorrect options:"
        )
        assert args.scoring_method_for_process_of_elimination == "calibration"
        assert (
            args.prompting_method_for_process_of_elimination
            == "multiple_choice_prompt"
        )
        assert args.mask_strategy_for_process_of_elimination == "min_k"
        assert args.do_synonym is True
        assert args.number_of_synonyms == 3
        assert (
            args.generate_synonyms_prompt
            == "Generate synonyms for option: option"
        )
        assert args.push_data_to_hub is True
        assert args.min_k == 2
        assert args.mask_token == "[MASK]"


def test_parse_args_missing_required_arguments():
    test_args = [
        "script_name",
        "--model_family",
        "GPT2",
        # "--checkpoint" is missing
        "--datasets",
        "copa",
    ]
    with mock.patch.object(sys, "argv", test_args):
        with pytest.raises(SystemExit):
            parse_args()


# Tests for load_data function
@pytest.mark.parametrize(
    "dataset_name,loader_name,ending_names,header_name",
    [
        ("copa", "copa_loader", ["hypothesis0", "hypothesis1"], "premise"),
        (
            "cqa",
            "cqa_loader",
            [
                "hypothesis0",
                "hypothesis1",
                "hypothesis2",
                "hypothesis3",
                "hypothesis4",
            ],
            "premise",
        ),
        (
            "obqa",
            "obqa_loader",
            ["hypothesis0", "hypothesis1", "hypothesis2", "hypothesis3"],
            "premise",
        ),
        ("piqa", "piqa_loader", ["hypothesis0", "hypothesis1"], "premise"),
        (
            "qasc",
            "qasc_loader",
            [
                "hypothesis0",
                "hypothesis1",
                "hypothesis2",
                "hypothesis3",
                "hypothesis4",
                "hypothesis5",
                "hypothesis6",
                "hypothesis7",
            ],
            "premise",
        ),
        (
            "siqa",
            "siqa_loader",
            ["hypothesis0", "hypothesis1", "hypothesis2"],
            "premise",
        ),
        (
            "winogrande",
            "winogrande_loader",
            ["hypothesis0", "hypothesis1"],
            "premise",
        ),
        (
            "anli",
            "anli_loader",
            ["hypothesis0", "hypothesis1", "hypothesis2"],
            "premise",
        ),
        (
            "disambiguation_qa",
            "date_understanding_loader",
            ["hypothesis0", "hypothesis1", "hypothesis2"],
            "premise",
        ),
        (
            "conceptual_combinations",
            "date_understanding_loader",
            ["hypothesis0", "hypothesis1", "hypothesis2", "hypothesis3"],
            "premise",
        ),
        (
            "date_understanding",
            "date_understanding_loader",
            [
                "hypothesis0",
                "hypothesis1",
                "hypothesis2",
                "hypothesis3",
                "hypothesis4",
                "hypothesis5",
            ],
            "premise",
        ),
        (
            "emoji_movie",
            "date_understanding_loader",
            [
                "hypothesis0",
                "hypothesis1",
                "hypothesis2",
                "hypothesis3",
                "hypothesis4",
            ],
            "premise",
        ),
        (
            "ruin_names",
            "date_understanding_loader",
            ["hypothesis0", "hypothesis1", "hypothesis2", "hypothesis3"],
            "premise",
        ),
        (
            "penguins_in_a_table",
            "date_understanding_loader",
            [
                "hypothesis0",
                "hypothesis1",
                "hypothesis2",
                "hypothesis3",
                "hypothesis4",
            ],
            "premise",
        ),
        (
            "strange_stories",
            "date_understanding_loader",
            ["hypothesis0", "hypothesis1", "hypothesis2", "hypothesis3"],
            "premise",
        ),
        (
            "reasoning_about_colored_objects",
            "date_understanding_loader",
            [f"hypothesis{i}" for i in range(18)],
            "premise",
        ),
        (
            "symbol_interpretation",
            "date_understanding_loader",
            [
                "hypothesis0",
                "hypothesis1",
                "hypothesis2",
                "hypothesis3",
                "hypothesis4",
            ],
            "premise",
        ),
        (
            "tracking_shuffled_objects",
            "date_understanding_loader",
            [
                "hypothesis0",
                "hypothesis1",
                "hypothesis2",
                "hypothesis3",
                "hypothesis4",
            ],
            "premise",
        ),
        (
            "logical_deduction_three_objects",
            "date_understanding_loader",
            ["hypothesis0", "hypothesis1", "hypothesis2"],
            "premise",
        ),
        (
            "logical_deduction_five_objects",
            "date_understanding_loader",
            [
                "hypothesis0",
                "hypothesis1",
                "hypothesis2",
                "hypothesis3",
                "hypothesis4",
            ],
            "premise",
        ),
        (
            "logical_deduction_seven_objects",
            "date_understanding_loader",
            [
                "hypothesis0",
                "hypothesis1",
                "hypothesis2",
                "hypothesis3",
                "hypothesis4",
                "hypothesis5",
                "hypothesis6",
            ],
            "premise",
        ),
        (
            "anli_r1",
            "anli_loader",
            ["hypothesis0", "hypothesis1", "hypothesis2"],
            "premise",
        ),
        (
            "vqa",
            "vqa_loader",
            [f"hypothesis{i}" for i in range(18)],
            "premise",
        ),
        (
            "scienceqa",
            "scienceqa_loader",
            [f"hypothesis{i}" for i in range(4)],
            "premise",
        ),
        (
            "ai2d",
            "ai2d_loader",
            [f"hypothesis{i}" for i in range(4)],
            "premise",
        ),
        (
            "single_inference",
            "single_inference_loader",
            [f"hypothesis{i}" for i in range(4)],
            "premise",
        ),
    ],
)
def test_load_data_datasets(
    dataset_name, loader_name, ending_names, header_name
):
    # Create a mock args object
    class Args:
        dataset = dataset_name
        image_path = None
        sample = None
        n_shot = 0
        num_options = len(ending_names)

    args = Args()

    # Mock the data loader function
    loader_path = f"mm_poe.methods.utils.utils.{loader_name}"
    with mock.patch(loader_path) as mock_loader:
        # Mock return value
        mock_value = {
            "premise": "Test premise",
            "uncond_premise": "Test premise",
            "image_path": "dummy_path",
            "label": 0,
        }
        for i, ending_name in enumerate(ending_names):
            mock_value[ending_name] = f"answer {i}"
        mock_loader.return_value = [mock_value]
        # Mock os.path.join to prevent file system access
        with mock.patch("os.path.join", return_value="dummy_path"):
            if dataset_name in [
                "vqa",
                "scienceqa",
                "ai2d",
                "single_inference",
            ]:
                ending, header, image_header, dev_dataset, train_dataset = (
                    load_data(args)
                )
                assert image_header == "image_path"
            else:
                ending, header, dev_dataset, train_dataset = load_data(args)
            assert ending == ending_names
            assert header == header_name
            assert len(dev_dataset) == 1
            assert len(train_dataset) == 1


def test_load_data_invalid_dataset():
    class Args:
        dataset = "unknown_dataset"

    args = Args()

    with mock.patch("builtins.print") as mock_print:
        result = load_data(args)
        assert result is None
        mock_print.assert_called_with(
            f"{args.dataset}: downloader not implemented."
        )


# Tests for load_model function
@pytest.mark.parametrize(
    "model_family,model_func_name,tokenizer_func_name",
    [
        ("GPT2", "AutoModelForCausalLM", "AutoTokenizer"),
        ("Pythia", "AutoModelForCausalLM", "AutoTokenizer"),
        ("OPT-IML", "AutoModelForCausalLM", "AutoTokenizer"),
        ("Dolly", "AutoModelForCausalLM", "AutoTokenizer"),
        ("T5", "AutoModelForSeq2SeqLM", "AutoTokenizer"),
        ("FLAN-T5", "AutoModelForSeq2SeqLM", "AutoTokenizer"),
        ("BLIP2-OPT", "AutoModelForVision2Seq", "AutoProcessor"),
        ("BLIP2-T5", "AutoModelForVision2Seq", "AutoProcessor"),
        ("InstructBLIP", "AutoModelForVision2Seq", "AutoProcessor"),
        ("GIT", "AutoModelForVision2Seq", "AutoProcessor"),
        ("PaliGemma", "AutoModelForVision2Seq", "AutoProcessor"),
        ("Idefics2", "AutoModelForVision2Seq", "AutoProcessor"),
    ],
)
def test_load_model_families(
    model_family, model_func_name, tokenizer_func_name
):
    device = "cpu"
    model_path = "some-model-path"

    # Create a mock args object
    class Args:

        model_family = ""
        loading_precision = "FP32"

        def __init__(self, model_family):
            self.model_family = model_family

    args = Args(model_family)

    # Mock the tokenizer and model loading functions
    with mock.patch(
        f"mm_poe.methods.utils.utils.{tokenizer_func_name}"
    ) as mock_tokenizer_class:
        with mock.patch(
            f"mm_poe.methods.utils.utils.{model_func_name}"
        ) as mock_model_class:
            mock_tokenizer = MagicMock()
            mock_model = MagicMock()
            mock_tokenizer_class.from_pretrained.return_value = mock_tokenizer
            mock_model_class.from_pretrained.return_value = mock_model

            # Set the return value of get_memory_footprint to a numeric value
            mock_model.get_memory_footprint.return_value = (
                2 * 1024**3
            )  # 2 GB in bytes

            model, tokenizer = load_model(device, model_path, args)

            # Check that the correct tokenizer and model are loaded
            if model_family == "Dolly":
                mock_tokenizer_class.from_pretrained.assert_called_with(
                    model_path, padding_side="left"
                )
            elif model_family == "Idefics2":
                mock_tokenizer_class.from_pretrained.assert_called_with(
                    model_path, do_image_splitting=False
                )
            else:
                mock_tokenizer_class.from_pretrained.assert_called_with(
                    model_path
                )


def test_load_model_invalid_family():
    device = "cpu"
    model_path = "some-model-path"

    # Create a mock args object
    class Args:
        model_family = "UnknownFamily"
        loading_precision = "FP32"

    args = Args()

    with mock.patch("builtins.print") as mock_print:
        result = load_model(device, model_path, args)
        assert result is None
        mock_print.assert_called_with(
            f"{args.model_family}: downloader not implemented."
        )


def test_load_model_loading_precision_int8():
    device = "cpu"
    model_path = "some-model-path"

    # Create a mock args object
    class Args:
        model_family = "GPT2"
        loading_precision = "INT8"

    args = Args()

    # Mock the tokenizer and model loading functions
    with mock.patch(
        "mm_poe.methods.utils.utils.AutoTokenizer"
    ) as mock_tokenizer_class:
        with mock.patch(
            "mm_poe.methods.utils.utils.AutoModelForCausalLM"
        ) as mock_model_class:
            with mock.patch(
                "mm_poe.methods.utils.utils.BitsAndBytesConfig"
            ) as mock_bnb_config_class:
                mock_tokenizer = MagicMock()
                mock_model = MagicMock()
                mock_bnb_config = MagicMock()
                mock_tokenizer_class.from_pretrained.return_value = (
                    mock_tokenizer
                )
                mock_model_class.from_pretrained.return_value = mock_model
                mock_bnb_config_class.return_value = mock_bnb_config

                # Set the return value of get_memory_footprint to a numeric value
                mock_model.get_memory_footprint.return_value = (
                    2 * 1024**3
                )  # 2 GB in bytes

                model, tokenizer = load_model(device, model_path, args)

                # Check that BitsAndBytesConfig is called correctly
                mock_bnb_config_class.assert_called_with(
                    load_in_8bit=True, llm_int8_threshold=200.0
                )
                # Check that model is loaded with quantization config
                mock_model_class.from_pretrained.assert_called_with(
                    model_path,
                    torch_dtype=torch.float16,
                    device_map=device,
                    quantization_config=mock_bnb_config,
                )


def test_load_model_loading_precision_int4():
    device = "cpu"
    model_path = "some-model-path"

    # Create a mock args object
    class Args:
        model_family = "GPT2"
        loading_precision = "INT4"

    args = Args()

    # Mock the tokenizer and model loading functions
    with mock.patch(
        "mm_poe.methods.utils.utils.AutoTokenizer"
    ) as mock_tokenizer_class:
        with mock.patch(
            "mm_poe.methods.utils.utils.AutoModelForCausalLM"
        ) as mock_model_class:
            with mock.patch(
                "mm_poe.methods.utils.utils.BitsAndBytesConfig"
            ) as mock_bnb_config_class:
                mock_tokenizer = MagicMock()
                mock_model = MagicMock()
                mock_bnb_config = MagicMock()
                mock_tokenizer_class.from_pretrained.return_value = (
                    mock_tokenizer
                )
                mock_model_class.from_pretrained.return_value = mock_model
                mock_bnb_config_class.return_value = mock_bnb_config

                # Set the return value of get_memory_footprint to a numeric value
                mock_model.get_memory_footprint.return_value = (
                    2 * 1024**3
                )  # 2 GB in bytes

                model, tokenizer = load_model(device, model_path, args)

                # Check that BitsAndBytesConfig is called correctly
                mock_bnb_config_class.assert_called_with(
                    load_in_4bit=True,
                    bnb_4bit_quant_type="nf4",
                    bnb_4bit_use_double_quant=True,
                    bnb_4bit_compute_dtype=torch.bfloat16,
                )
                # Check that model is loaded with quantization config
                mock_model_class.from_pretrained.assert_called_with(
                    model_path,
                    device_map=device,
                    quantization_config=mock_bnb_config,
                )


def test_load_model_loading_precision_fp16():
    device = "cpu"
    model_path = "some-model-path"

    # Create a mock args object
    class Args:
        model_family = "GPT2"
        loading_precision = "FP16"

    args = Args()

    # Mock the tokenizer and model loading functions
    with mock.patch(
        "mm_poe.methods.utils.utils.AutoTokenizer"
    ) as mock_tokenizer_class:
        with mock.patch(
            "mm_poe.methods.utils.utils.AutoModelForCausalLM"
        ) as mock_model_class:
            mock_tokenizer = MagicMock()
            mock_model = MagicMock()
            mock_bnb_config = MagicMock()
            mock_tokenizer_class.from_pretrained.return_value = mock_tokenizer
            mock_model_class.from_pretrained.return_value = mock_model

            # Set the return value of get_memory_footprint to a numeric value
            mock_model.get_memory_footprint.return_value = (
                2 * 1024**3
            )  # 2 GB in bytes

            model, tokenizer = load_model(device, model_path, args)

            # Check that model is loaded with quantization config
            mock_model_class.from_pretrained.assert_called_with(
                model_path, torch_dtype=torch.float16, device_map=device
            )


def test_load_model_loading_precision_bf16():
    device = "cpu"
    model_path = "some-model-path"

    # Create a mock args object
    class Args:
        model_family = "GPT2"
        loading_precision = "BF16"

    args = Args()

    # Mock the tokenizer and model loading functions
    with mock.patch(
        "mm_poe.methods.utils.utils.AutoTokenizer"
    ) as mock_tokenizer_class:
        with mock.patch(
            "mm_poe.methods.utils.utils.AutoModelForCausalLM"
        ) as mock_model_class:
            mock_tokenizer = MagicMock()
            mock_model = MagicMock()
            mock_bnb_config = MagicMock()
            mock_tokenizer_class.from_pretrained.return_value = mock_tokenizer
            mock_model_class.from_pretrained.return_value = mock_model

            # Set the return value of get_memory_footprint to a numeric value
            mock_model.get_memory_footprint.return_value = (
                2 * 1024**3
            )  # 2 GB in bytes

            model, tokenizer = load_model(device, model_path, args)

            # Check that model is loaded with quantization config
            mock_model_class.from_pretrained.assert_called_with(
                model_path, torch_dtype=torch.bfloat16, device_map=device
            )


# Tests for write_to_csv function
def test_write_to_csv_process_of_elimination(tmp_path):
    save_path = tmp_path / "results.csv"

    class Args:
        method = "process_of_elimination"
        model_family = "GPT2"
        checkpoint = "gpt2-medium"
        loading_precision = "FP32"
        dataset = "copa"
        batch_size = 32
        scoring_method_for_process_of_elimination = "language_modeling"
        prompting_method_for_process_of_elimination = "multiple_choice_prompt"
        mask_strategy_for_process_of_elimination = "lowest"
        mask_token = None
        seed = 42
        n_shot = 0
        sample = None
        mask_accuracy = 0.8

    args = Args()
    total_accuracy = 0.85

    write_to_csv(str(save_path), args, total_accuracy)

    with open(save_path, "r") as f:
        content = f.read()
        assert "process_of_elimination" in content
        assert f"{args.mask_accuracy:.4f}" in content
        assert f"{total_accuracy:.4f}" in content


def test_write_to_csv_contrastive_decoding(tmp_path):
    save_path = tmp_path / "results.csv"

    class Args:
        method = "contrastive_decoding"
        model_family = "GPT2"
        checkpoint = "gpt2-medium"
        amateur_checkpoint = "gpt2-small"
        loading_precision = "FP32"
        dataset = "copa"
        batch_size = 32
        expert_method = "language_modeling"
        amateur_method = "calibration"
        weighting_parameter = -1.0
        seed = 42
        n_shot = 0
        sample = None
        expert_accuracy = 0.9
        amateur_accuracy = 0.7

    args = Args()
    total_accuracy = 0.85

    write_to_csv(str(save_path), args, total_accuracy)

    with open(save_path, "r") as f:
        content = f.read()
        assert "contrastive_decoding" in content
        assert f"{args.expert_accuracy:.4f}" in content
        assert f"{args.amateur_accuracy:.4f}" in content
        assert f"{total_accuracy:.4f}" in content


def test_write_to_csv_generate_synonyms(tmp_path):
    save_path = tmp_path / "results.csv"

    class Args:
        method = "generate_synonyms"
        model_family = "GPT2"
        checkpoint = "gpt2-medium"
        loading_precision = "FP32"
        dataset = "copa"
        batch_size = 32
        number_of_synonyms = 5
        seed = 42
        n_shot = 0
        sample = None

    args = Args()
    total_accuracy = 0.88

    write_to_csv(str(save_path), args, total_accuracy)

    with open(save_path, "r") as f:
        content = f.read()
        assert "generate_synonyms" in content
        assert f"{total_accuracy:.4f}" in content


def test_write_to_csv_default_method(tmp_path):
    save_path = tmp_path / "results.csv"

    class Args:
        method = "default_method"
        model_family = "GPT2"
        checkpoint = "gpt2-medium"
        loading_precision = "FP32"
        dataset = "copa"
        batch_size = 32
        seed = 42
        n_shot = 0
        sample = None

    args = Args()
    total_accuracy = 0.9

    write_to_csv(str(save_path), args, total_accuracy)

    with open(save_path, "r") as f:
        content = f.read()
        assert "default_method" in content
        assert f"{total_accuracy:.4f}" in content


# Additional tests for branches and edge cases
def test_parse_args_invalid_choice():
    test_args = [
        "script_name",
        "--model_family",
        "GPT2",
        "--checkpoint",
        "gpt2-medium",
        "--datasets",
        "copa",
        "--loading_precision",
        "INVALID_PRECISION",
    ]
    with mock.patch.object(sys, "argv", test_args):
        with pytest.raises(SystemExit):
            parse_args()


def test_write_to_csv_no_method(tmp_path):
    save_path = tmp_path / "results.csv"

    class Args:
        method = None
        model_family = "GPT2"
        checkpoint = "gpt2-medium"
        loading_precision = "FP32"
        dataset = "copa"
        batch_size = 32
        seed = 42
        n_shot = 0
        sample = 100

    args = Args()
    total_accuracy = 0.9

    # with pytest.raises(AttributeError):
    write_to_csv(str(save_path), args, total_accuracy)
    assert os.path.isfile(save_path) == True



"""Model Downloaders"""

import argparse
import glob
import os
import shutil

import torch
from transformers import (
    AutoTokenizer,
    AutoModelForCausalLM,
    AutoModelForSeq2SeqLM,
    AutoProcessor,
    AutoModelForVision2Seq,
)

all_checkpoints = {
    "GPT2": ["gpt2", "gpt2-medium", "gpt2-large", "gpt2-xl"],
    "Pythia": [
        "EleutherAI/pythia-70m",
        "EleutherAI/pythia-160m",
        "EleutherAI/pythia-410m",
        "EleutherAI/pythia-1b",
        "EleutherAI/pythia-1.4b",
        "EleutherAI/pythia-2.8b",
        "EleutherAI/pythia-6.9B",
        "EleutherAI/pythia-12b",
        "EleutherAI/pythia-70m-deduped",
        "EleutherAI/pythia-160m-deduped",
        "EleutherAI/pythia-410m-deduped",
        "EleutherAI/pythia-1b-deduped",
        "EleutherAI/pythia-1.4b-deduped",
        "EleutherAI/pythia-2.8b-deduped",
        "EleutherAI/pythia-6.9B-deduped",
        "EleutherAI/pythia-12b-deduped",
    ],
    "OPT-IML": ["facebook/opt-iml-1.3b", "facebook/opt-iml-max-1.3b"],
    "T5": ["t5-small", "t5-base", "t5-large", "t5-3b", "t5-11b"],
    "FLAN-T5": [
        "google/flan-t5-small",
        "google/flan-t5-base",
        "google/flan-t5-large",
        "google/flan-t5-xl",
        "google/flan-t5-xxl",
    ],
    "MPT": [
        "mosaicml/mpt-7b",
        "mosaicml/mpt-7b-instruct",
        "mosaicml/mpt-7b-chat",
        "mosaicml/mpt-7b-storywriter",
    ],
    "Dolly": ["databricks/dolly-v2-7b"],
    "BLIP2-OPT": ["Salesforce/blip2-opt-2.7b"],
    "BLIP2-T5": ["Salesforce/blip2-flan-t5-xl"],
    "InstructBLIP": ["Salesforce/instructblip-vicuna-7b"],
    "GIT": ["microsoft/git-base-vqav2", "microsoft/git-base-textvqa"],
    "PaliGemma": [
        "google/paligemma-3b-ft-science-qa-448",
        "google/paligemma-3b-ft-vqav2-448",
        "google/paligemma-3b-ft-ai2d-448",
    ],
    "Idefics2": ["HuggingFaceM4/idefics2-8b"],
}


def parse_args():
    """Parse Arguments"""
    parser = argparse.ArgumentParser(description="Language model downloaders.")

    parser.add_argument(
        "--model_family",
        type=str,
        choices=[
            "GPT2",
            "T5",
            "FLAN-T5",
            "Pythia",
            "OPT-IML",
            "Dolly",
            "BLIP2-OPT",
            "BLIP2-T5",
            "InstructBLIP",
            "GIT",
            "PaliGemma",
            "Idefics2",
        ],
        default=None,
        help="The model family, as checkpoints under "
        + "the same model family use same codes to download.",
    )

    parser.add_argument(
        "--checkpoint",
        type=str,
        default=None,
        help="The checkpoint name under a model family, "
        + "e.g. gpt2, gpt2-medium, gpt2-large, gpt2-xl.",
    )

    parser.add_argument(
        "--download_all_checkpoints",
        action="store_true",
        help="If set to true, downlaod all checkpoitns of a model family.",
    )

    parser.add_argument(
        "--output_dir",
        type=str,
        default="./models",
    )

    args = parser.parse_args()
    return args


def main():
    """Main"""
    # import pdb; pdb.set_trace()
    args = parse_args()

    if args.model_family in ["GPT2", "Pythia", "OPT-IML", "Dolly"]:
        tokenizer_func = AutoTokenizer
        model_func = AutoModelForCausalLM
    elif args.model_family in ["T5", "FLAN-T5"]:
        tokenizer_func = AutoTokenizer
        model_func = AutoModelForSeq2SeqLM
    elif args.model_family in [
        "BLIP2-OPT",
        "BLIP2-T5",
        "InstructBLIP",
        "GIT",
        "PaliGemma",
        "Idefics2",
    ]:
        tokenizer_func = AutoProcessor
        model_func = AutoModelForVision2Seq
    else:
        print(f"{args.model_family}: downloader not implemented.")
        return

    # Check the validity of the checkpoint
    checkpoints = []
    if args.download_all_checkpoints is True:
        checkpoints = all_checkpoints[args.model_family]
    elif args.checkpoint not in all_checkpoints[args.model_family]:
        print(
            f"Invalid checkpoint from {args.model_family}. Choose from: "
            + f"{all_checkpoints[args.model_family]} or "
            + "set --download_all_checkpoints"
        )
        return
    else:
        checkpoints = [args.checkpoint]

    print(f"Models to download: {checkpoints}")
    for checkpoint in checkpoints:
        print(
            f"Downloading {checkpoint}\t under model family "
            + f"{args.model_family}..."
        )

        # download the model
        # tokenizer = tokenizer_func.from_pretrained(checkpoint)
        # model = model_func.from_pretrained(checkpoint)
        if args.model_family == "Dolly":
            tokenizer = tokenizer_func.from_pretrained(
                checkpoint, padding_side="left"
            )
            model = model_func.from_pretrained(
                checkpoint,
                torch_dtype=torch.bfloat16,
                device_map="cuda",
                # torch_dtype=torch.float16,
                # load_in_8bit=True,
            )
        elif args.model_family in [
            "BLIP2-OPT",
            "BLIP2-T5",
            "InstructBLIP",
            "PaliGemma",
            "Idefics2",
        ]:
            tokenizer = tokenizer_func.from_pretrained(checkpoint)
            model = model_func.from_pretrained(
                checkpoint, torch_dtype=torch.float16, device_map="cuda"
            )
        else:
            model = model_func.from_pretrained(checkpoint)
            tokenizer = tokenizer_func.from_pretrained(checkpoint)

        # save the model
        save_dir = os.path.join(args.output_dir, args.model_family, checkpoint)
        os.makedirs(os.path.dirname(save_dir), exist_ok=True)
        tokenizer.save_pretrained(save_dir)
        model.save_pretrained(save_dir)

        # delete cached files
        # https://huggingface.co/docs/transformers/installation#cache-setup
        cache_dir = "~/.cache/huggingface/hub"

        folders_to_delete = glob.glob(os.path.join(cache_dir, "models*"))
        for folder in folders_to_delete:
            print(f"Removing cached files at {folder}...")
            shutil.rmtree(folder)


if __name__ == "__main__":
    main()



# a framework for inference on multiple choice tasks.
import copy
import logging
import os

import torch
from torch.utils.data import DataLoader


from mm_poe.methods.utils.data import (
    upload_to_huggingface_hub,
    preprocess_function_seq2seq,
    preprocess_function_causal,
    preprocess_function_causal_channel,
    preprocess_function_seq2seq_channel,
    create_multiple_choice_prompt,
    generate_n_shot_poe_demonstrations,
    create_n_shot_splits,
)
from mm_poe.methods.utils.methods import (
    compute_conditional_score_seq2seq,
    compute_conditional_score_causal,
    compute_mask_process_of_elimination,
    inference_process_of_elimination,
    inference_language_modeling,
    inference_calibration,
)
from mm_poe.methods.utils.utils import (
    load_data,
    load_model,
    parse_args,
    set_seed,
    write_to_csv,
)

logger = logging.getLogger(__name__)


def main():
    # import pdb; pdb.set_trace()

    # step 1: argument parser, and logger
    args = parse_args()
    args.method = "process_of_elimination"

    # print(args)
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
    )
    logger.setLevel(logging.INFO)

    # step 2: set random seed to ensure reproducibility.
    logger.info(f"Set random seed to {args.seed}.")
    set_seed(args.seed)

    # step 3: load model, tokenizer.
    # Then move to gpu, and set to evaluation mode.
    logger.info(f"Load {args.model_family} model: {args.checkpoint}.")
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # get model path: ../models/args.model_family/args.checkpoint
    model_path = os.path.join("../models", args.model_family, args.checkpoint)
    model, tokenizer = load_model(device, model_path, args)
    if args.model_family in ["GPT2", "Pythia", "OPT-IML", "Dolly"]:
        compute_func = compute_conditional_score_causal
        preprocess_func = preprocess_function_causal
        preprocess_func_channel = preprocess_function_causal_channel
        remove_columns = ["input_ids", "labels", "ending_attention_mask"]
    elif args.model_family in ["T5", "FLAN-T5"]:
        compute_func = compute_conditional_score_seq2seq
        preprocess_func = preprocess_function_seq2seq
        preprocess_func_channel = preprocess_function_seq2seq_channel
        remove_columns = [
            "header_input_ids",
            "header_attention_mask",
            "ending_input_ids",
            "ending_attention_mask",
        ]
    else:
        raise NotImplementedError

    # step 4: load and preprocess data.
    args.datasets = args.datasets.split()
    logger.info(f"Load data: {args.datasets}.")

    # evaluate on each dataset
    multiple_choice_prompt = args.multiple_choice_prompt
    for dataset in args.datasets:
        args.dataset = dataset
        # multiple_choice_prompt = args.multiple_choice_prompt
        args.multiple_choice_prompt = None
        ending_names, header_name, raw_dataset, n_shot_dataset = load_data(
            args
        )
        raw_dataset, n_shot_dataset, n_shot_demonstrations = (
            create_n_shot_splits(raw_dataset, n_shot_dataset, args)
        )

        mcp_args = copy.deepcopy(args)
        mcp_args.multiple_choice_prompt = multiple_choice_prompt
        _, _, raw_mcp_dataset, n_shot_mcp_dataset = load_data(mcp_args)
        raw_mcp_dataset, n_shot_mcp_dataset, _ = create_n_shot_splits(
            raw_mcp_dataset, n_shot_mcp_dataset, args
        )

        logger.info(f"Preprocess data: {args.dataset}.")
        fn_kwargs = {
            "ending_names": ending_names,
            "header_name": header_name,
            "tokenizer": tokenizer,
        }
        num_of_options = len(ending_names)
        tokenized_dataset = raw_dataset.map(
            preprocess_func,
            fn_kwargs=fn_kwargs,
            batched=True,
            batch_size=args.batch_size,
        )
        eval_dataloader = DataLoader(
            tokenized_dataset, batch_size=args.batch_size, shuffle=False
        )

        # step 5: (evaluation) inference on data, and compute accuracy.
        logger.info(
            f"Start inference (method: {args.method}) on {args.dataset} "
            + f"using {args.model_family} model: {args.checkpoint}."
        )
        scoring_method = args.scoring_method_for_process_of_elimination
        logger.info(
            f"Step 1: Computing masks. Scoring method: {scoring_method}."
        )
        if scoring_method == "channel":
            tokenized_channel_dataset = raw_dataset.map(
                preprocess_func_channel,
                fn_kwargs=fn_kwargs,
                batched=True,
                batch_size=args.batch_size,
            )
            eval_channel_dataloader = DataLoader(
                tokenized_channel_dataset,
                batch_size=args.batch_size,
                shuffle=False,
            )
            avg_log_probs, _, _, _ = inference_language_modeling(
                model,
                eval_channel_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        elif scoring_method == "calibration":
            fn_kwargs = {
                "ending_names": ending_names,
                "header_name": "uncond_premise",  # the difference is here
                "tokenizer": tokenizer,
            }
            tokenized_calibration_dataset = raw_dataset.map(
                preprocess_func,
                fn_kwargs=fn_kwargs,
                batched=True,
                batch_size=args.batch_size,
            )
            eval_calibration_dataloader = DataLoader(
                tokenized_calibration_dataset,
                batch_size=args.batch_size,
                shuffle=False,
            )
            avg_log_probs, _, _, _ = inference_calibration(
                model,
                eval_dataloader,
                eval_calibration_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        elif scoring_method == "language_modeling":
            avg_log_probs, _, _, _ = inference_language_modeling(
                model,
                eval_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        elif scoring_method == "multiple_choice_prompt":
            # mcp_args = copy.deepcopy(args)
            # mcp_args.multiple_choice_prompt = multiple_choice_prompt
            # _, _, raw_mcp_dataset, n_shot_mcp_dataset = load_data(mcp_args)
            # raw_mcp_dataset, n_shot_mcp_dataset = create_n_shot_splits(
            # raw_mcp_dataset, n_shot_mcp_dataset, args)
            tokenized_dataset = raw_mcp_dataset.map(
                preprocess_func,
                fn_kwargs=fn_kwargs,
                batched=True,
                batch_size=args.batch_size,
            )
            eval_mcp_dataloader = DataLoader(
                tokenized_dataset, batch_size=args.batch_size, shuffle=False
            )
            avg_log_probs, _, _, _ = inference_language_modeling(
                model,
                eval_mcp_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        else:
            raise NotImplementedError  # unlikely to happen.

        mask_strategy = args.mask_strategy_for_process_of_elimination
        if mask_strategy == "min_k":
            # masking the most k UNLIKELY options
            min_k = args.min_k
            if min_k >= num_of_options:
                min_k = num_of_options - 1
            mask_kwargs = {
                "min_k": min_k,
            }
        else:
            mask_kwargs = {}
        masks = compute_mask_process_of_elimination(
            avg_log_probs, mask_strategy, **mask_kwargs
        )
        # construct an oracle mask that only keeps the correct lable to 1,
        # and other options to 0
        # oracle_masks = torch.zeros_like(avg_log_probs)
        # oracle_masks[torch.arange(oracle_masks.size(0)), \
        # tokenized_dataset["label"]] = 1
        masks = masks.to(torch.float32)
        # compute mask accuracy, i.e.,
        # check whether mask that correspond to labels is 1
        mask_result = masks[
            torch.arange(masks.size(0)), tokenized_dataset["label"]
        ]
        mask_accuracy = torch.sum(mask_result) / mask_result.size(0)
        logger.info(f"Mask accuracy: {mask_accuracy}")
        args.mask_accuracy = mask_accuracy.item()
        masked_dataset = tokenized_dataset.map(
            lambda example, idx: {"mask": masks[idx]},
            with_indices=True,
            batched=True,
            remove_columns=remove_columns,
        )

        prompting_method = args.prompting_method_for_process_of_elimination
        logger.info(
            "Step 2: Creating multiple choice prompt. "
            + f"Prompting method: {prompting_method}."
        )
        # if args.prompting_method_for_process_of_elimination
        # mcp_kwargs = {"multiple_choice_prompt": multiple_choice_prompt,}
        mask_token = args.mask_token
        if mask_token is not None:
            if mask_token == "":
                args.process_of_elimination_prompt = (
                    args.process_of_elimination_prompt.replace(
                        "[MASK]", "empty"
                    )
                )
            else:
                args.process_of_elimination_prompt = (
                    args.process_of_elimination_prompt.replace(
                        "[MASK]", mask_token
                    )
                )
        mcp_kwargs = {
            "multiple_choice_prompt": args.process_of_elimination_prompt,
            "scoring_method": scoring_method,
            "num_of_options": num_of_options,
            "mask_token": mask_token,
        }
        mcp_dataset = masked_dataset.map(
            create_multiple_choice_prompt, fn_kwargs=mcp_kwargs
        )
        # change n_shot format.
        if args.n_shot > 0:
            if (
                args.scoring_method_for_process_of_elimination
                == "multiple_choice_prompt"
            ):
                n_shot_demonstrations, n_shot_poe_demonstrations = (
                    generate_n_shot_poe_demonstrations(
                        n_shot_mcp_dataset, num_of_options
                    )
                )
            else:
                _, n_shot_poe_demonstrations = (
                    generate_n_shot_poe_demonstrations(
                        n_shot_mcp_dataset, num_of_options
                    )
                )
            mcp_dataset = mcp_dataset.map(
                lambda x: {
                    "premise": x["premise"].replace(
                        n_shot_demonstrations, n_shot_poe_demonstrations
                    )
                }
            )

        logger.info("Step 3: Final Inference")
        mcp_dataset = mcp_dataset.map(
            preprocess_func,
            fn_kwargs=fn_kwargs,
            batched=True,
            batch_size=args.batch_size,
        )
        eval_mcp_dataloader = DataLoader(
            mcp_dataset, batch_size=args.batch_size, shuffle=False
        )
        poe_avg_log_probs, lm_accuracy, _, _ = (
            inference_process_of_elimination(
                model,
                eval_mcp_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        )

        # step 6: some postprocessing, including saving and displyaing output.
        save_path = os.path.join("../results", f"{args.method}.csv")
        logger.info(f"Save results to {save_path}.")
        save_args = copy.deepcopy(args)
        if mask_strategy == "min_k":
            save_args.mask_strategy_for_process_of_elimination = (
                f"min_k_{min_k}"
            )
        write_to_csv(save_path, save_args, lm_accuracy)

        # step 7: push data to HuggingFace Hub.
        if args.push_data_to_hub:
            logger.info(f"Push {args.dataset} to HuggingFace Hub.")
            # save the mcp dataset, which will be used by LLM.
            upload_to_huggingface_hub(mcp_dataset, args)


if __name__ == "__main__":
    main()



# a framework for inference on multiple choice tasks.
from argparse import Namespace
import copy
import logging
import os
import pathlib
import sys

import questionary
import torch
from torch.utils.data import DataLoader

from mm_poe.methods.utils.data import (
    create_multiple_choice_prompt,
    preprocess_function_seq2seq_vqa,
    preprocess_function_seq2seq_vqa_channel,
    preprocess_function_causal_vqa,
    preprocess_function_causal_vqa_channel,
)
from mm_poe.methods.utils.methods import (
    compute_conditional_score_seq2seq_vqa,
    compute_conditional_score_causal_vqa,
    compute_mask_process_of_elimination,
    inference_process_of_elimination,
    inference_language_modeling,
    inference_calibration,
)
from mm_poe.methods.utils.utils import load_data, load_model, set_seed

from mm_poe.models.model_downloaders.model_downloaders import (
    main as download_model,
)

all_checkpoints = {
    "BLIP2-OPT": ["Salesforce/blip2-opt-2.7b"],
    "BLIP2-T5": ["Salesforce/blip2-flan-t5-xl"],
    "InstructBLIP": ["Salesforce/instructblip-vicuna-7b"],
    "GIT": ["microsoft/git-base-vqav2", "microsoft/git-base-textvqa"],
    "PaliGemma": [
        "google/paligemma-3b-ft-science-qa-448",
        "google/paligemma-3b-ft-vqav2-448",
        "google/paligemma-3b-ft-ai2d-448",
    ],
    "Idefics2": ["HuggingFaceM4/idefics2-8b"],
}

logger = logging.getLogger(__name__)


def main():
    """
    The main function executes on commands:
    `python -m mm_poe` and `$ mm_poe `.
    """
    # step 1: collect arguments
    args = Namespace()
    args.seed = 0

    args.model_family = questionary.select(
        message="Select model family?",
        choices=[
            "BLIP2-OPT",
            "BLIP2-T5",
            "InstructBLIP",
            "GIT",
            "PaliGemma",
            "Idefics2",
        ],
        default="GIT",
    ).ask()

    checkpoints_choices = all_checkpoints[args.model_family]
    args.checkpoint = questionary.select(
        message="Select model checkpoint?",
        choices=checkpoints_choices,
        default=checkpoints_choices[0],
    ).ask()

    args.loading_precision = questionary.select(
        message="Select model precision?",
        choices=["FP32", "FP16", "BF16", "INT8", "INT4"],
        default="FP32",
    ).ask()

    args.output_dir = questionary.path(
        message="Model output directory?",
        only_directories=True,
        default="./models/",
    ).ask()

    args.dataset = "single_inference"
    args.batch_size = 1
    args.sample = 1
    args.n_shot = 0

    args.multiple_choice_prompt = ""
    args.calibration_prompt = " the answer is:"
    args.process_of_elimination_prompt = (
        "Select the most suitable option "
        + "to answer the question. Ignore [MASK] options."
    )

    args.scoring_method_for_process_of_elimination = questionary.select(
        message="Select scoring method?",
        choices=[
            "channel",
            "calibration",
            "language_modeling",
            "multiple_choice_prompt",
        ],
        default="language_modeling",
    ).ask()

    args.mask_strategy_for_process_of_elimination = questionary.select(
        message="Select mask strategy?",
        choices=["below_average", "lowest"],
        default="below_average",
    ).ask()

    args.prompting_method_for_process_of_elimination = "multiple_choice_prompt"
    args.mask_token = None

    args.question = questionary.text("Question:").ask().strip()
    args.choices = questionary.text("Choices [comma seprated]:").ask()
    args.choices = [choice.strip() for choice in args.choices.split(",")]
    args.num_options = len(args.choices)
    args.image_path = questionary.path(
        "Image Path?", default="./images/image.png"
    ).ask()
    args.label = questionary.select(
        message="Ground Truth Option:",
        choices=[str(x) for x in range(args.num_options)],
    ).ask()
    args.label = int(args.label)
    args.method = "process_of_elimination"

    # print(args)
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
    )
    logger.setLevel(logging.INFO)

    # step 2: set random seed to ensure reproducibility.
    logger.info(f"Set random seed to {args.seed}.")
    set_seed(args.seed)

    # step 3: download model
    logger.info(f"Download {args.model_family} model: {args.checkpoint}.")
    model_downloader_path = os.path.join(
        pathlib.Path(__file__).parent.resolve(),
        "models/model_downloaders/model_downloaders.py",
    )
    sys.argv = [
        model_downloader_path,
        "--model_family",
        args.model_family,
        "--checkpoint",
        args.checkpoint,
        "--output_dir",
        args.output_dir,
    ]
    download_model()

    # step 4: load model, tokenizer.
    # Then move to gpu, and set to evaluation mode.
    logger.info(f"Load {args.model_family} model: {args.checkpoint}.")
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # get model path: ../models/args.model_family/args.checkpoint
    model_path = os.path.join(
        args.output_dir, args.model_family, args.checkpoint
    )
    model, tokenizer = load_model(device, model_path, args)
    if args.model_family in [
        "BLIP2-T5",
        "InstructBLIP",
        "PaliGemma",
        "Idefics2",
    ]:
        compute_func = compute_conditional_score_seq2seq_vqa
        preprocess_func = preprocess_function_seq2seq_vqa
        preprocess_func_channel = preprocess_function_seq2seq_vqa_channel
        remove_columns = [
            "header_input_ids",
            "header_attention_mask",
            "ending_input_ids",
            "ending_attention_mask",
            "images",
        ]
        processor = tokenizer
        tokenizer = processor.tokenizer
    elif args.model_family in ["BLIP2-OPT", "GIT"]:
        compute_func = compute_conditional_score_causal_vqa
        preprocess_func = preprocess_function_causal_vqa
        preprocess_func_channel = preprocess_function_causal_vqa_channel
        remove_columns = [
            "input_ids",
            "labels",
            "images",
            "ending_attention_mask",
        ]
        processor = tokenizer
        tokenizer = processor.tokenizer
    else:
        raise NotImplementedError

    # step 5: load and preprocess data.
    logger.info(f"Load data: {args.dataset}.")

    # evaluate on dataset
    multiple_choice_prompt = args.multiple_choice_prompt
    args.multiple_choice_prompt = None
    (
        ending_names,
        header_name,
        image_header_name,
        raw_dataset,
        n_shot_dataset,
    ) = load_data(args)

    mcp_args = copy.deepcopy(args)
    mcp_args.multiple_choice_prompt = multiple_choice_prompt
    _, _, _, raw_mcp_dataset, n_shot_mcp_dataset = load_data(mcp_args)

    logger.info(f"Preprocess data: {args.dataset}.")
    fn_kwargs = {
        "ending_names": ending_names,
        "header_name": header_name,
        "tokenizer": tokenizer,
        "processor": processor,
        "image_header_name": image_header_name,
        "image_token": "<image>",
    }
    if args.model_family in ["GIT"]:
        fn_kwargs["image_token"] = ""
    num_of_options = len(ending_names)
    tokenized_dataset = raw_dataset.map(
        preprocess_func,
        fn_kwargs=fn_kwargs,
        batched=True,
        batch_size=args.batch_size,
    )
    eval_dataloader = DataLoader(
        tokenized_dataset, batch_size=args.batch_size, shuffle=False
    )

    # step 5: (evaluation) inference on data, and compute accuracy.
    logger.info(
        f"Start inference (method: {args.method}) on {args.dataset} "
        + f"using {args.model_family} model: {args.checkpoint}."
    )
    scoring_method = args.scoring_method_for_process_of_elimination
    logger.info(f"Step 1: Computing masks. Scoring method: {scoring_method}.")
    if scoring_method == "channel":
        tokenized_channel_dataset = raw_dataset.map(
            preprocess_func_channel,
            fn_kwargs=fn_kwargs,
            batched=True,
            batch_size=args.batch_size,
        )
        eval_channel_dataloader = DataLoader(
            tokenized_channel_dataset,
            batch_size=args.batch_size,
            shuffle=False,
        )
        avg_log_probs, _, _, lm_predictions = inference_language_modeling(
            model,
            eval_channel_dataloader,
            device,
            compute_func,
            tokenizer.pad_token_id,
        )
    elif scoring_method == "calibration":
        fn_kwargs = {
            "ending_names": ending_names,
            "header_name": "uncond_premise",  # the difference is here
            "tokenizer": tokenizer,
            "processor": processor,
            "image_header_name": image_header_name,
            "image_token": "<image>",
        }
        if args.model_family in ["GIT"]:
            fn_kwargs["image_token"] = ""
        tokenized_calibration_dataset = raw_dataset.map(
            preprocess_func,
            fn_kwargs=fn_kwargs,
            batched=True,
            batch_size=args.batch_size,
        )
        eval_calibration_dataloader = DataLoader(
            tokenized_calibration_dataset,
            batch_size=args.batch_size,
            shuffle=False,
        )
        avg_log_probs, _, _, lm_predictions = inference_calibration(
            model,
            eval_dataloader,
            eval_calibration_dataloader,
            device,
            compute_func,
            tokenizer.pad_token_id,
        )
    elif scoring_method == "language_modeling":
        avg_log_probs, _, _, lm_predictions = inference_language_modeling(
            model,
            eval_dataloader,
            device,
            compute_func,
            tokenizer.pad_token_id,
        )
    elif scoring_method == "multiple_choice_prompt":
        # mcp_args = copy.deepcopy(args)
        # mcp_args.multiple_choice_prompt = multiple_choice_prompt
        # _, _, raw_mcp_dataset, n_shot_mcp_dataset = load_data(mcp_args)
        # raw_mcp_dataset, n_shot_mcp_dataset = create_n_shot_splits(
        #     raw_mcp_dataset,
        #     n_shot_mcp_dataset,
        #     args
        # )
        tokenized_dataset = raw_mcp_dataset.map(
            preprocess_func,
            fn_kwargs=fn_kwargs,
            batched=True,
            batch_size=args.batch_size,
        )
        eval_mcp_dataloader = DataLoader(
            tokenized_dataset, batch_size=args.batch_size, shuffle=False
        )
        avg_log_probs, _, _, lm_predictions = inference_language_modeling(
            model,
            eval_mcp_dataloader,
            device,
            compute_func,
            tokenizer.pad_token_id,
        )
    else:
        raise NotImplementedError  # unlikely to happen.

    mask_strategy = args.mask_strategy_for_process_of_elimination
    if mask_strategy == "min_k":
        # masking the most k UNLIKELY options
        min_k = args.min_k
        if min_k >= num_of_options:
            min_k = num_of_options - 1
        mask_kwargs = {
            "min_k": min_k,
        }
    else:
        mask_kwargs = {}
    masks = compute_mask_process_of_elimination(
        avg_log_probs, mask_strategy, **mask_kwargs
    )
    # construct an oracle mask that only keeps the correct lable to 1,
    # and other options to 0
    # oracle_masks = torch.zeros_like(avg_log_probs)
    # oracle_masks[torch.arange(oracle_masks.size(0)), \
    #              tokenized_dataset["label"]] = 1
    masks = masks.to(torch.float32)
    # compute mask accuracy, i.e.,
    # check whether mask that correspond to labels is 1
    mask_result = masks[
        torch.arange(masks.size(0)), tokenized_dataset["label"]
    ]
    mask_accuracy = torch.sum(mask_result) / mask_result.size(0)
    logger.info(f"Mask accuracy: {mask_accuracy}")
    args.mask_accuracy = mask_accuracy.item()
    masked_dataset = tokenized_dataset.map(
        lambda example, idx: {"mask": masks[idx]},
        with_indices=True,
        batched=True,
        remove_columns=remove_columns,
    )

    prompting_method = args.prompting_method_for_process_of_elimination
    logger.info(
        "Step 2: Creating multiple choice prompt. "
        + f"Prompting method: {prompting_method}."
    )
    # if args.prompting_method_for_process_of_elimination
    # mcp_kwargs = {"multiple_choice_prompt": multiple_choice_prompt,}
    mask_token = args.mask_token
    if mask_token is not None:
        if mask_token == "":
            args.process_of_elimination_prompt = (
                args.process_of_elimination_prompt.replace("[MASK]", "empty")
            )
        else:
            args.process_of_elimination_prompt = (
                args.process_of_elimination_prompt.replace(
                    "[MASK]", mask_token
                )
            )
    mcp_kwargs = {
        "multiple_choice_prompt": args.process_of_elimination_prompt,
        "scoring_method": scoring_method,
        "num_of_options": num_of_options,
        "mask_token": mask_token,
    }
    mcp_dataset = masked_dataset.map(
        create_multiple_choice_prompt, fn_kwargs=mcp_kwargs
    )

    logger.info("Step 3: Final Inference")
    mcp_dataset = mcp_dataset.map(
        preprocess_func,
        fn_kwargs=fn_kwargs,
        batched=True,
        batch_size=args.batch_size,
    )
    eval_mcp_dataloader = DataLoader(
        mcp_dataset, batch_size=args.batch_size, shuffle=False
    )
    poe_avg_log_probs, lm_accuracy, _, lm_predictions = (
        inference_process_of_elimination(
            model,
            eval_mcp_dataloader,
            device,
            compute_func,
            tokenizer.pad_token_id,
        )
    )
    option = int(lm_predictions.numpy()[0])
    logger.info(f"Predicted Option: {option}. Answer: {args.choices[option]}")



# test_main.py

import pytest
from unittest.mock import patch
import runpy


def test_main_called():
    with patch("mm_poe.cli.main") as mock_main:
        # Simulate running __main__.py as the main module
        runpy.run_module("mm_poe.__main__", run_name="__main__")
        mock_main.assert_called_once()


def test_main_not_called_when_imported():
    with patch("mm_poe.cli.main") as mock_main:
        # Import __main__.py as a module; __name__ will not be '__main__'
        import mm_poe.__main__

        mock_main.assert_not_called()



import os
import sys
import torch
import pytest
from unittest import mock
from unittest.mock import MagicMock, patch
from torch.utils.data import DataLoader
from tqdm import tqdm

# Import the functions from methods.py
from mm_poe.methods.utils.methods import (
    inference_language_modeling_old,
    inference_contrastive_decoding_old,
    inference_language_modeling,
    inference_generate_synonyms,
    inference_calibration,
    inference_contrastive_decoding,
    compute_mask_process_of_elimination,
    inference_process_of_elimination,
    compute_conditional_score_seq2seq,
    compute_conditional_score_causal,
    compute_conditional_score_seq2seq_vqa,
    compute_conditional_score_causal_vqa,
    generate_synonyms,
    aggregate_optionw_with_synonyms,
)

# Mock tqdm to prevent actual progress bars during testing
tqdm = lambda x, **kwargs: x


# Define a simple mock model
class SimpleMockModel(torch.nn.Module):
    def __init__(self):
        super(SimpleMockModel, self).__init__()

    def forward(
        self, input_ids=None, labels=None, pixel_values=None, **kwargs
    ):
        seq_len = input_ids.size(1)
        batch_size_num_options = labels.size(0)
        vocab_size = 32128
        logits = torch.randn(batch_size_num_options, seq_len, vocab_size)
        loss = torch.tensor(0.0)
        return MagicMock(loss=loss, logits=logits)


# Fixtures for common test components
@pytest.fixture
def mock_model():
    return SimpleMockModel()


@pytest.fixture
def mock_amateur_model():
    return SimpleMockModel()


@pytest.fixture
def mock_expert_model():
    return SimpleMockModel()


@pytest.fixture
def sample_batch():
    batch_size = 2
    num_options = 2
    seq_len = 18
    vocab_size = 32128
    return {
        "ending_input_ids": torch.randint(
            0, vocab_size, (batch_size, num_options, seq_len)
        ),
        "header_input_ids": torch.randint(
            0, vocab_size, (batch_size, seq_len)
        ),
        "label": torch.randint(0, num_options, (batch_size,)),
        "header_attention_mask": torch.ones(batch_size, seq_len),
        "ending_attention_mask": torch.ones(batch_size, num_options, seq_len),
        "input_ids": torch.randint(
            0, vocab_size, (batch_size, num_options, seq_len)
        ),
        "labels": torch.randint(
            0, vocab_size, (batch_size, num_options, seq_len)
        ),
        "mask": torch.ones(batch_size, num_options),
        "images": torch.randn(batch_size, 3, 224, 224),
    }


@pytest.fixture
def sample_synonym_batch():
    batch_size = 2
    num_synonyms = 2
    num_options = 2
    seq_len = 18
    vocab_size = 32128
    return {
        "ending_input_ids": torch.randint(
            0,
            vocab_size,
            (batch_size, num_options * (num_synonyms + 1), seq_len),
        ),
        "header_input_ids": torch.randint(
            0, vocab_size, (batch_size, seq_len)
        ),
        "label": torch.randint(
            0, num_options * (num_synonyms + 1), (batch_size,)
        ),
        "header_attention_mask": torch.ones(batch_size, seq_len),
        "ending_attention_mask": torch.ones(
            batch_size, num_options * (num_synonyms + 1), seq_len
        ),
        "input_ids": torch.randint(
            0,
            vocab_size,
            (batch_size, num_options * (num_synonyms + 1), seq_len),
        ),
        "labels": torch.randint(
            0,
            vocab_size,
            (batch_size, num_options * (num_synonyms + 1), seq_len),
        ),
        "mask": torch.ones(batch_size, num_options * (num_synonyms + 1)),
        "images": torch.randn(batch_size, 3, 224, 224),
    }


@pytest.fixture
def device():
    return "cpu"


@pytest.fixture
def pad_token_id():
    return 0


# Tests for inference_language_modeling_old function
def test_inference_language_modeling_old(mock_model, sample_batch, device):
    eval_dataloader = [sample_batch]
    total_accuracy = inference_language_modeling_old(
        mock_model, eval_dataloader, device
    )
    assert isinstance(total_accuracy, float)
    assert 0.0 <= total_accuracy <= 1.0


# Tests for inference_contrastive_decoding_old function
def test_inference_contrastive_decoding_old(
    mock_amateur_model, mock_expert_model, sample_batch, device
):
    eval_dataloader = [sample_batch]
    total_accuracy = inference_contrastive_decoding_old(
        mock_amateur_model, mock_expert_model, eval_dataloader, device
    )
    assert isinstance(total_accuracy, float)
    assert 0.0 <= total_accuracy <= 1.0


# Mock compute_func for inference_language_modeling
def mock_compute_func(batch, model, device, pad_token_id):
    batch_size = batch["header_input_ids"].size(0)
    num_options = batch["ending_input_ids"].size(1)
    return torch.rand(batch_size, num_options)


# Tests for inference_language_modeling function
def test_inference_language_modeling(
    mock_model, sample_batch, device, pad_token_id
):
    eval_dataloader = [sample_batch]
    avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
        inference_language_modeling(
            mock_model,
            eval_dataloader,
            device,
            mock_compute_func,
            pad_token_id,
        )
    )
    assert avg_log_probs.shape == (
        sample_batch["label"].size(0),
        sample_batch["ending_input_ids"].size(1),
    )
    assert isinstance(lm_accuracy, float)
    assert isinstance(avg_lm_accuracy, float)
    assert lm_predictions.shape == (sample_batch["label"].size(0),)


# Tests for inference_generate_synonyms function
def test_inference_generate_synonyms(
    mock_model, sample_synonym_batch, device, pad_token_id
):
    num_of_options = 2
    num_of_synonyms = 2

    def mock_compute_func(batch, model, device, pad_token_id):
        batch_size = batch["header_input_ids"].size(0)
        total_options = batch["ending_input_ids"].size(1)
        return torch.rand(batch_size, total_options)

    eval_dataloader = [sample_synonym_batch]
    avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
        inference_generate_synonyms(
            mock_model,
            eval_dataloader,
            device,
            mock_compute_func,
            pad_token_id,
            num_of_options,
            num_of_synonyms,
        )
    )
    expected_shape = (
        sample_synonym_batch["label"].size(0),
        num_of_options * (num_of_synonyms + 1),
    )
    assert avg_log_probs.shape == expected_shape
    assert isinstance(lm_accuracy, float)
    assert isinstance(avg_lm_accuracy, float)
    assert lm_predictions.shape == (sample_synonym_batch["label"].size(0),)


# Tests for inference_calibration function
def test_inference_calibration(mock_model, sample_batch, device, pad_token_id):
    eval_dataloader = [sample_batch]
    eval_calibration_dataloader = [sample_batch]
    avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
        inference_calibration(
            mock_model,
            eval_dataloader,
            eval_calibration_dataloader,
            device,
            mock_compute_func,
            pad_token_id,
        )
    )
    assert avg_log_probs.shape == (
        sample_batch["label"].size(0),
        sample_batch["ending_input_ids"].size(1),
    )
    assert isinstance(lm_accuracy, float)
    assert isinstance(avg_lm_accuracy, float)
    assert lm_predictions.shape == (sample_batch["label"].size(0),)


# Tests for compute_mask_process_of_elimination function
@pytest.mark.parametrize(
    "mask_strategy", ["lowest", "below_average", "lowest_iter", "min_k"]
)
def test_compute_mask_process_of_elimination(mask_strategy):
    avg_log_probs = torch.tensor([[0.1, 0.2, 0.3], [0.3, 0.2, 0.1]])
    if mask_strategy == "min_k":
        kwargs = {"min_k": 2}
    else:
        kwargs = {}
    if mask_strategy not in [
        "lowest",
        "below_average",
        "lowest_iter",
        "min_k",
    ]:
        with pytest.raises(NotImplementedError):
            compute_mask_process_of_elimination(
                avg_log_probs, mask_strategy, **kwargs
            )
    else:
        masks = compute_mask_process_of_elimination(
            avg_log_probs, mask_strategy, **kwargs
        )
        assert masks.shape == avg_log_probs.shape


# Tests for inference_process_of_elimination function
def test_inference_process_of_elimination(
    mock_model, sample_batch, device, pad_token_id
):
    eval_dataloader = [sample_batch]
    avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
        inference_process_of_elimination(
            mock_model,
            eval_dataloader,
            device,
            mock_compute_func,
            pad_token_id,
        )
    )
    assert avg_log_probs.shape == (
        sample_batch["label"].size(0),
        sample_batch["ending_input_ids"].size(1),
    )
    assert isinstance(lm_accuracy, float)
    assert isinstance(avg_lm_accuracy, float)
    assert lm_predictions.shape == (sample_batch["label"].size(0),)


# Tests for compute_conditional_score_seq2seq function
def test_compute_conditional_score_seq2seq(
    mock_model, sample_batch, device, pad_token_id
):
    log_prob = compute_conditional_score_seq2seq(
        sample_batch, mock_model, device, pad_token_id
    )
    assert log_prob.shape == (
        sample_batch["ending_input_ids"].shape[0],
        sample_batch["ending_input_ids"].shape[1],
    )


# Tests for compute_conditional_score_causal function
def test_compute_conditional_score_causal(
    mock_model, sample_batch, device, pad_token_id
):
    log_prob = compute_conditional_score_causal(
        sample_batch, mock_model, device, pad_token_id
    )
    assert log_prob.shape == (
        sample_batch["input_ids"].shape[0],
        sample_batch["input_ids"].shape[1],
    )


# Tests for compute_conditional_score_seq2seq_vqa function
def test_compute_conditional_score_seq2seq_vqa(
    mock_model, sample_batch, device, pad_token_id
):
    log_prob = compute_conditional_score_seq2seq_vqa(
        sample_batch, mock_model, device, pad_token_id
    )
    assert log_prob.shape == (
        sample_batch["ending_input_ids"].shape[0],
        sample_batch["ending_input_ids"].shape[1],
    )


# Tests for compute_conditional_score_causal_vqa function
def test_compute_conditional_score_causal_vqa(
    mock_model, sample_batch, device, pad_token_id
):
    log_prob = compute_conditional_score_causal_vqa(
        sample_batch, mock_model, device, pad_token_id
    )
    assert log_prob.shape == (
        sample_batch["input_ids"].shape[0],
        sample_batch["input_ids"].shape[1],
    )


# Tests for aggregate_optionw_with_synonyms function
def test_aggregate_optionw_with_synonyms():
    batch_size = 2
    num_of_options = 5
    num_of_synonyms = 3
    tensor = torch.arange(
        batch_size * num_of_options * (num_of_synonyms + 1)
    ).view(batch_size, -1)
    aggregated_tensor = aggregate_optionw_with_synonyms(
        tensor.clone(), num_of_options, num_of_synonyms
    )
    assert aggregated_tensor.shape == tensor.shape


# Tests for generate_synonyms function
def test_generate_synonyms():
    args = MagicMock()
    args.number_of_synonyms = 2
    args.generate_synonyms_prompt = "Generate a synonym to '{option}':"
    model = MagicMock()
    model.device = "cpu"
    tokenizer = MagicMock()
    tokenizer.return_tensors = "pt"
    tokenizer.pad_token_id = 0
    tokenizer.batch_decode.return_value = ["synonym1", "synonym2"]
    tokenized_dataset = MagicMock()
    tokenized_dataset.column_names = ["hypothesis1"]
    tokenized_dataset.__getitem__.return_value = {"hypothesis1": "test_option"}
    synonyms_dict = generate_synonyms(
        args, model, tokenizer, tokenized_dataset
    )
    assert isinstance(synonyms_dict, dict)


# Tests for inference_contrastive_decoding function
def test_inference_contrastive_decoding():
    method = "language_modeling"
    model = MagicMock()
    args = MagicMock()
    args.batch_size = 2
    args.model_family = "other"
    raw_dataset = MagicMock()
    device = "cpu"
    compute_func = MagicMock()
    tokenizer = MagicMock()
    processor = MagicMock()
    ending_names = ["ending1", "ending2"]
    header_name = "header"
    image_header_name = "image_header"
    preprocess_func = MagicMock()
    preprocess_func_channel = MagicMock()
    kwargs = {
        "args": args,
        "raw_dataset": raw_dataset,
        "device": device,
        "compute_func": compute_func,
        "tokenizer": tokenizer,
        "processor": processor,
        "ending_names": ending_names,
        "header_name": header_name,
        "image_header_name": image_header_name,
        "preprocess_func": preprocess_func,
        "preprocess_func_channel": preprocess_func_channel,
    }
    with patch(
        "mm_poe.methods.utils.methods.inference_language_modeling",
        return_value=(None, 0.0, 0.0, None),
    ) as mock_inference:
        avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
            inference_contrastive_decoding(method, model, **kwargs)
        )
        mock_inference.assert_called_once()

    method = "calibration"
    with patch(
        "mm_poe.methods.utils.methods.inference_calibration",
        return_value=(None, 0.0, 0.0, None),
    ) as mock_inference_cal:
        avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
            inference_contrastive_decoding(method, model, **kwargs)
        )
        mock_inference_cal.assert_called_once()

    method = "channel"
    with patch(
        "mm_poe.methods.utils.methods.inference_language_modeling",
        return_value=(None, 0.0, 0.0, None),
    ) as mock_inference_channel:
        avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
            inference_contrastive_decoding(method, model, **kwargs)
        )
        mock_inference_channel.assert_called()

    method = "invalid_method"
    with pytest.raises(NotImplementedError):
        inference_contrastive_decoding(method, model, **kwargs)



import json
import os
import xml.etree.ElementTree as ET
import random

import torch
from PIL import Image

# write my own data loader, or using HF dataloader?
# steps for data loader: label, premise, options, hypothesis.
# uncond_premise = " the answer is:"


def upload_to_huggingface_hub(dataset, args):
    suffix = (
        f"{args.dataset}_{args.seed}_{args.n_shot}_"
        + f"{args.sample}_{args.checkpoint.split('/')[-1]}_{args.batch_size}"
    )
    temp_data_path = os.path.join(f"../temp_data/{args.method}", suffix)
    dataset.save_to_disk(temp_data_path)
    # _ = upload_folder(
    #     folder_path=temp_data_path,
    #     path_in_repo=f"temp_data/{args.method}/{suffix}",
    #     repo_id="Vanmas/PoE_data",
    #     repo_type="dataset",)
    # remove the temp data folder
    # os.system(f"rm -rf {temp_data_path}")


def preprocess_function_seq2seq(examples, **kwargs):
    ending_names, header_name, tokenizer = (
        kwargs["ending_names"],
        kwargs["header_name"],
        kwargs["tokenizer"],
    )
    num_choice = len(ending_names)
    question_headers = examples[header_name]
    # the tokenizer handles multiple spaces.
    first_sentences = [
        [context] * len(ending_names) for context in examples[header_name]
    ]
    # second_sentences = [
    #     [f"{header} {examples[end][i]}" for end in ending_names] \
    # for i, header in enumerate(question_header)
    # ]
    second_sentences = [
        [f"{examples[end][i]}" for end in ending_names]
        for i, header in enumerate(question_headers)
    ]

    first_sentences = sum(first_sentences, [])
    second_sentences = sum(second_sentences, [])

    # tokenized_examples = tokenizer(first_sentences, \
    # second_sentences, truncation=True)
    tokenized_headers = tokenizer(
        first_sentences, padding=True, truncation=True
    )
    tokenized_endings = tokenizer(
        second_sentences, padding=True, truncation=True
    )
    header_dict = {
        f"header_{k}": [
            v[i : i + num_choice] for i in range(0, len(v), num_choice)
        ]
        for k, v in tokenized_headers.items()
    }
    ending_dict = {
        f"ending_{k}": [
            v[i : i + num_choice] for i in range(0, len(v), num_choice)
        ]
        for k, v in tokenized_endings.items()
    }
    return {**header_dict, **ending_dict}


def preprocess_function_causal(examples, **kwargs):
    ending_names, header_name, tokenizer = (
        kwargs["ending_names"],
        kwargs["header_name"],
        kwargs["tokenizer"],
    )
    num_choice = len(ending_names)
    question_headers = examples[header_name]
    # the tokenizer handles multiple spaces.
    first_sentences = [
        [context] * len(ending_names) for context in examples[header_name]
    ]
    # second_sentences = [
    #     [f"{header} {examples[end][i]}" for end in ending_names] \
    # for i, header in enumerate(question_header)
    # ]
    second_sentences = [
        [f"{examples[end][i]}" for end in ending_names]
        for i, header in enumerate(question_headers)
    ]

    first_sentences = sum(first_sentences, [])
    second_sentences = sum(second_sentences, [])

    # tokenized_examples = tokenizer(first_sentences, \
    # second_sentences, truncation=True)
    tokenized_headers = tokenizer(first_sentences, truncation=True)
    tokenized_endings = tokenizer(second_sentences, truncation=True)

    max_len = max(
        len(header + ending)
        for header, ending in zip(
            tokenized_headers["input_ids"], tokenized_endings["input_ids"]
        )
    )
    input_ids = torch.full(
        (len(tokenized_headers["input_ids"]), max_len),
        tokenizer.pad_token_id,
        dtype=torch.long,
    )
    labels = tokenizer.pad_token_id * torch.ones(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    ending_attention_mask = torch.zeros(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    for i, (header, ending) in enumerate(
        zip(tokenized_headers["input_ids"], tokenized_endings["input_ids"])
    ):
        input_ids[i, : len(header)] = torch.tensor(header)
        input_ids[i, len(header) : len(header) + len(ending)] = torch.tensor(
            ending
        )
        ending_attention_mask[i, len(header) : len(header) + len(ending)] = (
            torch.tensor(1)
        )
        labels[i, len(header) : len(header) + len(ending)] = torch.tensor(
            ending
        )

    flatten_dict = {
        "input_ids": input_ids,
        "labels": labels,
        "ending_attention_mask": ending_attention_mask,
    }
    return_dict = {
        f"{k}": [v[i : i + num_choice] for i in range(0, len(v), num_choice)]
        for k, v in flatten_dict.items()
    }
    return return_dict


def preprocess_function_seq2seq_vqa(examples, **kwargs):
    ending_names, header_name, image_header_name, processor, image_token = (
        kwargs["ending_names"],
        kwargs["header_name"],
        kwargs["image_header_name"],
        kwargs["processor"],
        kwargs["image_token"],
    )
    tokenizer = processor.tokenizer
    image_processor = processor.image_processor

    num_choice = len(ending_names)
    question_headers = examples[header_name]
    # the tokenizer handles multiple spaces.
    first_sentences = [
        [image_token + context] * len(ending_names)
        for context in examples[header_name]
    ]
    second_sentences = [
        [f"{examples[end][i]}" for end in ending_names]
        for i, header in enumerate(question_headers)
    ]

    first_sentences = sum(first_sentences, [])
    second_sentences = sum(second_sentences, [])

    # tokenized_examples = tokenizer(first_sentences, \
    # second_sentences, truncation=True)
    tokenized_headers = tokenizer(
        first_sentences, padding=True, truncation=True
    )
    tokenized_endings = tokenizer(
        second_sentences, padding=True, truncation=True
    )

    image_paths = examples[image_header_name]
    images = [
        Image.open(image_path).convert("RGB") for image_path in image_paths
    ]
    images = [[image] * len(ending_names) for image in images]
    images = sum(images, [])
    images = image_processor(images, return_tensors="pt").data

    flatten_image_dict = {"images": images["pixel_values"]}
    header_dict = {
        f"header_{k}": [
            v[i : i + num_choice] for i in range(0, len(v), num_choice)
        ]
        for k, v in tokenized_headers.items()
    }
    ending_dict = {
        f"ending_{k}": [
            v[i : i + num_choice] for i in range(0, len(v), num_choice)
        ]
        for k, v in tokenized_endings.items()
    }
    image_dict = {
        f"{k}": [v[i : i + num_choice] for i in range(0, len(v), num_choice)]
        for k, v in flatten_image_dict.items()
    }
    return {**header_dict, **ending_dict, **image_dict}


def preprocess_function_causal_vqa(examples, **kwargs):
    ending_names, header_name, image_header_name, processor, image_token = (
        kwargs["ending_names"],
        kwargs["header_name"],
        kwargs["image_header_name"],
        kwargs["processor"],
        kwargs["image_token"],
    )
    tokenizer = processor.tokenizer
    image_processor = processor.image_processor

    num_choice = len(ending_names)
    question_headers = examples[header_name]
    first_sentences = [
        [image_token + context] * len(ending_names)
        for context in examples[header_name]
    ]
    second_sentences = [
        [f"{examples[end][i]}" for end in ending_names]
        for i, header in enumerate(question_headers)
    ]

    first_sentences = sum(first_sentences, [])
    second_sentences = sum(second_sentences, [])

    # print(first_sentences)
    # print(second_sentences)

    tokenized_headers = tokenizer(first_sentences, truncation=True)
    tokenized_endings = tokenizer(second_sentences, truncation=True)

    image_paths = examples[image_header_name]
    images = [
        Image.open(image_path).convert("RGB") for image_path in image_paths
    ]
    images = [[image] * len(ending_names) for image in images]
    images = sum(images, [])
    images = image_processor(images, return_tensors="pt").data

    max_len = max(
        len(header + ending)
        for header, ending in zip(
            tokenized_headers["input_ids"], tokenized_endings["input_ids"]
        )
    )
    input_ids = torch.full(
        (len(tokenized_headers["input_ids"]), max_len),
        tokenizer.pad_token_id,
        dtype=torch.long,
    )
    labels = tokenizer.pad_token_id * torch.ones(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    header_attention_mask = torch.zeros(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    ending_attention_mask = torch.zeros(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    for i, (header, ending) in enumerate(
        zip(tokenized_headers["input_ids"], tokenized_endings["input_ids"])
    ):
        if tokenizer.padding_side == "right":
            input_ids[i, : len(header)] = torch.tensor(header)
            input_ids[i, len(header) : len(header) + len(ending)] = (
                torch.tensor(ending)
            )
            header_attention_mask[i, : len(header) + len(ending)] = (
                torch.tensor(1)
            )
            ending_attention_mask[
                i, len(header) : len(header) + len(ending)
            ] = torch.tensor(1)
            labels[i, len(header) : len(header) + len(ending)] = torch.tensor(
                ending
            )
        else:
            input_ids[i, -len(ending) :] = torch.tensor(ending)
            input_ids[i, -len(header) - len(ending) : -len(ending)] = (
                torch.tensor(header)
            )
            header_attention_mask[i, -len(header) - len(ending) :] = (
                torch.tensor(1)
            )
            ending_attention_mask[i, -len(ending) :] = torch.tensor(1)
            labels[i, -len(ending) :] = torch.tensor(ending)

    flatten_dict = {
        "input_ids": input_ids,
        "labels": labels,
        "header_attention_mask": header_attention_mask,
        "ending_attention_mask": ending_attention_mask,
        "images": images["pixel_values"],
    }
    return_dict = {
        f"{k}": [v[i : i + num_choice] for i in range(0, len(v), num_choice)]
        for k, v in flatten_dict.items()
    }
    return return_dict


def preprocess_function_seq2seq_channel(examples, **kwargs):
    ending_names, header_name, tokenizer = (
        kwargs["ending_names"],
        kwargs["header_name"],
        kwargs["tokenizer"],
    )
    num_choice = len(ending_names)
    question_headers = examples[header_name]
    # the tokenizer handles multiple spaces.
    first_sentences = [
        [context] * len(ending_names) for context in examples[header_name]
    ]
    # second_sentences = [
    #     [f"{header} {examples[end][i]}" for end in ending_names] \
    # for i, header in enumerate(question_header)
    # ]
    second_sentences = [
        [f"{examples[end][i]}" for end in ending_names]
        for i, header in enumerate(question_headers)
    ]

    first_sentences = sum(first_sentences, [])
    second_sentences = sum(second_sentences, [])

    # swap first_sentences and second_sentences
    first_sentences, second_sentences = second_sentences, first_sentences

    # tokenized_examples = tokenizer(first_sentences, \
    # second_sentences, truncation=True)
    tokenized_headers = tokenizer(
        first_sentences, padding=True, truncation=True
    )
    tokenized_endings = tokenizer(
        second_sentences, padding=True, truncation=True
    )
    header_dict = {
        f"header_{k}": [
            v[i : i + num_choice] for i in range(0, len(v), num_choice)
        ]
        for k, v in tokenized_headers.items()
    }
    ending_dict = {
        f"ending_{k}": [
            v[i : i + num_choice] for i in range(0, len(v), num_choice)
        ]
        for k, v in tokenized_endings.items()
    }
    return {**header_dict, **ending_dict}


def preprocess_function_causal_channel(examples, **kwargs):
    ending_names, header_name, tokenizer = (
        kwargs["ending_names"],
        kwargs["header_name"],
        kwargs["tokenizer"],
    )
    num_choice = len(ending_names)
    question_headers = examples[header_name]
    # the tokenizer handles multiple spaces.
    first_sentences = [
        [context] * len(ending_names) for context in examples[header_name]
    ]
    # second_sentences = [
    #     [f"{header} {examples[end][i]}" for end in ending_names] \
    # for i, header in enumerate(question_header)
    # ]
    second_sentences = [
        [f"{examples[end][i]}" for end in ending_names]
        for i, header in enumerate(question_headers)
    ]

    first_sentences = sum(first_sentences, [])
    second_sentences = sum(second_sentences, [])

    # swap first_sentences and second_sentences
    first_sentences, second_sentences = second_sentences, first_sentences

    # tokenized_examples = tokenizer(first_sentences, \
    # second_sentences, truncation=True)
    tokenized_headers = tokenizer(first_sentences, truncation=True)
    tokenized_endings = tokenizer(second_sentences, truncation=True)

    max_len = max(
        len(header + ending)
        for header, ending in zip(
            tokenized_headers["input_ids"], tokenized_endings["input_ids"]
        )
    )
    input_ids = torch.full(
        (len(tokenized_headers["input_ids"]), max_len),
        tokenizer.pad_token_id,
        dtype=torch.long,
    )
    labels = tokenizer.pad_token_id * torch.ones(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    ending_attention_mask = torch.zeros(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    for i, (header, ending) in enumerate(
        zip(tokenized_headers["input_ids"], tokenized_endings["input_ids"])
    ):
        input_ids[i, : len(header)] = torch.tensor(header)
        input_ids[i, len(header) : len(header) + len(ending)] = torch.tensor(
            ending
        )
        ending_attention_mask[i, len(header) : len(header) + len(ending)] = (
            torch.tensor(1)
        )
        labels[i, len(header) : len(header) + len(ending)] = torch.tensor(
            ending
        )

    flatten_dict = {
        "input_ids": input_ids,
        "labels": labels,
        "ending_attention_mask": ending_attention_mask,
    }
    return_dict = {
        f"{k}": [v[i : i + num_choice] for i in range(0, len(v), num_choice)]
        for k, v in flatten_dict.items()
    }
    return return_dict


def preprocess_function_seq2seq_vqa_channel(examples, **kwargs):
    ending_names, header_name, image_header_name, processor, image_token = (
        kwargs["ending_names"],
        kwargs["header_name"],
        kwargs["image_header_name"],
        kwargs["processor"],
        kwargs["image_token"],
    )
    tokenizer = processor.tokenizer
    image_processor = processor.image_processor

    num_choice = len(ending_names)
    question_headers = examples[header_name]
    # the tokenizer handles multiple spaces.
    first_sentences = [
        [context] * len(ending_names) for context in examples[header_name]
    ]
    second_sentences = [
        [image_token + f"{examples[end][i]}" for end in ending_names]
        for i, header in enumerate(question_headers)
    ]

    first_sentences = sum(first_sentences, [])
    second_sentences = sum(second_sentences, [])

    # swap first_sentences and second_sentences
    first_sentences, second_sentences = second_sentences, first_sentences

    # tokenized_examples = tokenizer(first_sentences, \
    # second_sentences, truncation=True)
    tokenized_headers = tokenizer(
        first_sentences, padding=True, truncation=True
    )
    tokenized_endings = tokenizer(
        second_sentences, padding=True, truncation=True
    )

    image_paths = examples[image_header_name]
    images = [
        Image.open(image_path).convert("RGB") for image_path in image_paths
    ]
    images = [[image] * len(ending_names) for image in images]
    images = sum(images, [])
    images = image_processor(images, return_tensors="pt").data

    flatten_image_dict = {"images": images["pixel_values"]}
    header_dict = {
        f"header_{k}": [
            v[i : i + num_choice] for i in range(0, len(v), num_choice)
        ]
        for k, v in tokenized_headers.items()
    }
    ending_dict = {
        f"ending_{k}": [
            v[i : i + num_choice] for i in range(0, len(v), num_choice)
        ]
        for k, v in tokenized_endings.items()
    }
    image_dict = {
        f"{k}": [v[i : i + num_choice] for i in range(0, len(v), num_choice)]
        for k, v in flatten_image_dict.items()
    }
    return {**header_dict, **ending_dict, **image_dict}


def preprocess_function_causal_vqa_channel(examples, **kwargs):
    ending_names, header_name, image_header_name, processor, image_token = (
        kwargs["ending_names"],
        kwargs["header_name"],
        kwargs["image_header_name"],
        kwargs["processor"],
        kwargs["image_token"],
    )
    tokenizer = processor.tokenizer
    image_processor = processor.image_processor

    num_choice = len(ending_names)
    question_headers = examples[header_name]
    first_sentences = [
        [image_token + context] * len(ending_names)
        for context in examples[header_name]
    ]
    second_sentences = [
        [f"{examples[end][i]}" for end in ending_names]
        for i, header in enumerate(question_headers)
    ]

    first_sentences = sum(first_sentences, [])
    second_sentences = sum(second_sentences, [])

    # swap first_sentences and second_sentences
    first_sentences, second_sentences = second_sentences, first_sentences

    tokenized_headers = tokenizer(first_sentences, truncation=True)
    tokenized_endings = tokenizer(second_sentences, truncation=True)

    image_paths = examples[image_header_name]
    images = [
        Image.open(image_path).convert("RGB") for image_path in image_paths
    ]
    images = [[image] * len(ending_names) for image in images]
    images = sum(images, [])
    images = image_processor(images, return_tensors="pt").data

    max_len = max(
        len(header + ending)
        for header, ending in zip(
            tokenized_headers["input_ids"], tokenized_endings["input_ids"]
        )
    )
    input_ids = torch.full(
        (len(tokenized_headers["input_ids"]), max_len),
        tokenizer.pad_token_id,
        dtype=torch.long,
    )
    labels = tokenizer.pad_token_id * torch.ones(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    header_attention_mask = torch.zeros(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    ending_attention_mask = torch.zeros(
        (len(tokenized_headers["input_ids"]), max_len), dtype=torch.long
    )
    for i, (header, ending) in enumerate(
        zip(tokenized_headers["input_ids"], tokenized_endings["input_ids"])
    ):
        if tokenizer.padding_side == "right":
            input_ids[i, : len(header)] = torch.tensor(header)
            input_ids[i, len(header) : len(header) + len(ending)] = (
                torch.tensor(ending)
            )
            header_attention_mask[i, : len(header) + len(ending)] = (
                torch.tensor(1)
            )
            ending_attention_mask[
                i, len(header) : len(header) + len(ending)
            ] = torch.tensor(1)
            labels[i, len(header) : len(header) + len(ending)] = torch.tensor(
                ending
            )
        else:
            input_ids[i, -len(ending) :] = torch.tensor(ending)
            input_ids[i, -len(header) - len(ending) : -len(ending)] = (
                torch.tensor(header)
            )
            header_attention_mask[i, -len(header) - len(ending) :] = (
                torch.tensor(1)
            )
            ending_attention_mask[i, -len(ending) :] = torch.tensor(1)
            labels[i, -len(ending) :] = torch.tensor(ending)

    flatten_dict = {
        "input_ids": input_ids,
        "labels": labels,
        "header_attention_mask": header_attention_mask,
        "ending_attention_mask": ending_attention_mask,
        "images": images["pixel_values"],
    }
    return_dict = {
        f"{k}": [v[i : i + num_choice] for i in range(0, len(v), num_choice)]
        for k, v in flatten_dict.items()
    }
    return return_dict


def create_multiple_choice_prompt(example, **kwargs):
    alphabets = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    multiple_choice_prompt = kwargs["multiple_choice_prompt"]
    scoring_method = kwargs["scoring_method"]
    num_of_options = kwargs["num_of_options"]
    mask = example["mask"]
    # null_string = f"[MASK]"
    if kwargs["mask_token"] is not None:
        null_string = kwargs["mask_token"]
    else:
        null_string = "[MASK]"
    mcp_example = {}
    # example['premise'] = premise = f"{multiple_choice_prompt} \
    # {premise}\nA. {options[0]}\nB. {options[1]}\nC. \
    # {options[2]}\nD. {options[3]}\nE. {options[4]}\nAnswer:"
    # premise = f"{multiple_choice_prompt} Question: {example['premise']}\n"

    if scoring_method != "multiple_choice_prompt":
        premise = (
            f"{multiple_choice_prompt}\n Question: {example['premise']}\n"
        )
        premise = premise.replace(f"{example['uncond_premise']}", "")
        for idx, single_mask in enumerate(mask):
            mcp_example[f"hypothesis{idx}"] = alphabets[idx]
            if single_mask == 1:
                premise += f"{alphabets[idx]}. {example[f'hypothesis{idx}']}\n"
            else:
                # consider other null strings.
                premise += f"{alphabets[idx]}. {null_string}\n"
        premise += "Answer:"
    else:
        # for multiple choice prompt,
        # options are already presented in the premise.
        premise = f"{multiple_choice_prompt}\n{example['premise']}"
        premise = premise.replace(f"{example['uncond_premise']}", "")
        for idx, single_mask in enumerate(mask):
            option_start_index = premise.rfind(f"{alphabets[idx]}. ")
            if idx == num_of_options - 1:
                option_end_index = premise.rfind("Answer:")
            else:
                option_end_index = premise.rfind(f"{alphabets[idx + 1]}. ")
            option = premise[option_start_index:option_end_index]
            if single_mask == 1:
                pass
            else:
                # consider other null strings.
                premise = premise.replace(
                    option, f"{alphabets[idx]}. {null_string}\n"
                )
    mcp_example["premise"] = premise
    return mcp_example


def create_synonym_dataset(examples, **kwargs):
    # for hypothesis0, create synonyms00, synonyms01, etc.
    args, synonyms_dict = kwargs["args"], kwargs["synonyms_dict"]
    number_of_synonyms = args.number_of_synonyms
    # get the hypothesis columns
    hypothesis_columns = [
        col for col in examples.keys() if "hypothesis" in col
    ]
    for hypothesis_column in hypothesis_columns:
        for i in range(number_of_synonyms):
            examples[f"{hypothesis_column}_synonyms_{i}"] = [
                synonyms_dict[hypothesis][i]
                for hypothesis in examples[hypothesis_column]
            ]
    return examples


def copa_loader(path, args):

    root = ET.parse(path).getroot()
    examples_copa = []
    for type_tag in root.findall("item"):
        # xml stuff
        value = type_tag.get("most-plausible-alternative")
        asks_for = type_tag.get("asks-for")
        children = list(type_tag)
        # get the texts
        p = children[0].text
        a1 = children[1].text[:1].lower() + children[1].text[1:]
        a2 = children[2].text[:1].lower() + children[2].text[1:]
        if asks_for == "effect":
            bridge = " so"
        elif asks_for == "cause":
            bridge = " because"
        else:
            assert False
        # examples_copa  += [{'options':
        #   [{'premise': ' ' + p[:-1] + bridge,
        #     'hypothesis': ' ' + a1,
        #     'uncond_premise': bridge,
        #     'uncond_hypothesis': ' ' + a1},
        # {'premise': ' ' + p[:-1] + bridge,
        #     'hypothesis': ' ' + a2,
        #     'uncond_premise': bridge,
        #     'uncond_hypothesis': ' ' + a2}],
        #           'label':int(value)-1}]
        premise = " " + p[:-1] + bridge
        if getattr(args, "multiple_choice_prompt", None) is not None:
            # Question: The pond froze over for the winter so
            # A. People skated on the pond.
            # B. People brought boats to the pond.
            # Answer:
            hypotheses = ["A", "B"]
            premise = (
                f"{args.multiple_choice_prompt} "
                + f"Question: {premise}\nA. {a1}\nB. {a2}\nAnswer:"
            )
        else:
            hypotheses = [" " + a1, " " + a2]
        examples_copa += [
            {
                "label": int(value) - 1,
                "premise": premise,
                "uncond_premise": bridge,
                "hypothesis0": hypotheses[0],
                "hypothesis1": hypotheses[1],
            }
        ]
    return examples_copa


def cqa_loader(path, args):
    examples_cqa = []
    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"
    with open(path) as f:
        for line in f:
            d = json.loads(line)
            label = ["A", "B", "C", "D", "E"].index(d["answerKey"])
            premise = " " + d["question"]["stem"]
            premise = premise[:-1] + "?"

            # to ensure identical perforamnce to the PMI paper.
            # options = [option['text'].lower() for option in \
            # d['question']['choices']]
            options = [
                f" \"{option['text'].lower()}\""
                for option in d["question"]["choices"]
            ]

            # examples += [{'options':
            # [{'premise':premise + '? the answer is:' ,
            #                 'hypothesis': ' "{}"'.format(c['text'].lower()),
            #                 'uncond_premise': ' the answer is:',
            #                 'uncond_hypothesis': \
            #   ' "{}"'.format(c['text'].lower())} \
            #   for c in d['question']['choices']],
            # 'label':label}]
            # if args.multiple_choice_prompt is not None:
            if getattr(args, "multiple_choice_prompt", None) is not None:
                hypotheses = ["A", "B", "C", "D", "E"]
                # Question: How does a bishop move from one place to another?
                # A. chess game
                # B. church
                # C. in a car
                # D. queen
                # E. cathedral
                # Answer:
                premise = (
                    f"{args.multiple_choice_prompt} "
                    + f"Question: {premise}\nA. {options[0]}\nB. "
                    + f"{options[1]}\nC. {options[2]}\nD. {options[3]}\nE. "
                    + f"{options[4]}\nAnswer:"
                )
            else:
                hypotheses = options
                premise = premise + uncond_premise
            examples_cqa += [
                {
                    "label": label,
                    "premise": premise,
                    "uncond_premise": uncond_premise,
                    "hypothesis0": hypotheses[0],
                    "hypothesis1": hypotheses[1],
                    "hypothesis2": hypotheses[2],
                    "hypothesis3": hypotheses[3],
                    "hypothesis4": hypotheses[4],
                }
            ]
    return examples_cqa


def obqa_loader(path, args):
    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"
    with open(path) as lines:
        abc2idx = {"A": 0, "B": 1, "C": 2, "D": 3}

        examples_obqa = []
        for line in lines:
            j = json.loads(line)

            label = abc2idx[j["answerKey"]]
            premise = j["question"]["stem"]
            options_text = [
                f" {option['text']}" for option in j["question"]["choices"]
            ]
            options_sym = [
                option["label"] for option in j["question"]["choices"]
            ]
            if getattr(args, "multiple_choice_prompt", None) is not None:
                # Question: Greenhouses are great for plants like
                # A. Pizza
                # B. Lollipops
                # C. Candles
                # D. French beans
                # Answer:
                hypotheses = options_sym
                premise = (
                    f"{args.multiple_choice_prompt} "
                    + f"Question: {premise}\nA. {options_text[0]}\nB. "
                    + f"{options_text[1]}\nC. {options_text[2]}\nD. "
                    + f"{options_text[3]}\nAnswer:"
                )
            else:
                hypotheses = options_text
                # premise = premise + uncond_premise

            examples_obqa += [
                {
                    "label": label,
                    "premise": premise,
                    "uncond_premise": uncond_premise,
                    "hypothesis0": hypotheses[0],
                    "hypothesis1": hypotheses[1],
                    "hypothesis2": hypotheses[2],
                    "hypothesis3": hypotheses[3],
                }
            ]
    return examples_obqa


def piqa_loader(path, args):
    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"
    examples_piqa = []
    qa_path, label_path = path[0], path[1]

    with open(qa_path) as lines, open(label_path) as labels:
        for line, label_sym in zip(lines, labels):
            label = int(label_sym[0])
            line = json.loads(line)
            premise = line["goal"]
            options_text = [line["sol1"], line["sol2"]]
            options_sym = ["A", "B"]

            if getattr(args, "multiple_choice_prompt", None) is not None:
                # Question: To clear snot out of your nose,
                # A. place a tissue over your nose and blow the snot out.
                # B. place a tissue over your nose and suck the snot in.
                # Answer:
                hypotheses = options_sym
                premise = (
                    f"{args.multiple_choice_prompt} "
                    + f"Question: {premise}\nA. {options_text[0]}\nB. "
                    + f"{options_text[1]}\nAnswer:"
                )
            else:
                hypotheses = options_text
                premise = premise + uncond_premise

            examples_piqa += [
                {
                    "label": label,
                    "premise": premise,
                    "uncond_premise": uncond_premise,
                    "hypothesis0": hypotheses[0],
                    "hypothesis1": hypotheses[1],
                }
            ]
    return examples_piqa


def qasc_loader(path, args):
    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"
    examples_qasc = []

    with open(path) as lines:
        for line in lines:
            line = json.loads(line)
            label = ["A", "B", "C", "D", "E", "F", "G", "H"].index(
                line["answerKey"]
            )
            premise = f"{line['question']['stem']}"

            options_text = [
                option["text"] for option in line["question"]["choices"]
            ]
            options_sym = [
                option["label"] for option in line["question"]["choices"]
            ]

            if getattr(args, "multiple_choice_prompt", None) is not None:
                # Question: Cameron returned home with a bag of
                # candy to eat all night
                # long. What will Others want to do next?
                # A. great
                # B. buy the candy to eat
                # C. bored
                # E. ...
                # Answer:
                hypotheses = options_sym
                premise = (
                    f"{args.multiple_choice_prompt} "
                    + f"Question: {premise}\nA. {options_text[0]}\nB. "
                    + f"{options_text[1]}\nC. {options_text[2]}\nD. "
                    + f"{options_text[3]}\nE. {options_text[4]}\nF. "
                    + f"{options_text[5]}\nG. {options_text[6]}\nH. "
                    + f"{options_text[7]}\nAnswer:"
                )
            else:
                hypotheses = options_text
                premise = premise + uncond_premise

            examples_qasc += [
                {
                    "label": label,
                    "premise": premise,
                    "uncond_premise": uncond_premise,
                    "hypothesis0": hypotheses[0],
                    "hypothesis1": hypotheses[1],
                    "hypothesis2": hypotheses[2],
                    "hypothesis3": hypotheses[3],
                    "hypothesis4": hypotheses[4],
                    "hypothesis5": hypotheses[5],
                    "hypothesis6": hypotheses[6],
                    "hypothesis7": hypotheses[7],
                }
            ]
    return examples_qasc


def siqa_loader(path, args):
    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"
    examples_siqa = []
    qa_path, label_path = path[0], path[1]

    with open(qa_path) as lines, open(label_path) as labels:
        for line, label_sym in zip(lines, labels):
            label = int(label_sym[0]) - 1
            line = json.loads(line)
            premise = f"{line['context']} {line['question']}"

            options_text = [line["answerA"], line["answerB"], line["answerC"]]
            options_sym = ["A", "B", "C"]

            if getattr(args, "multiple_choice_prompt", None) is not None:
                # Question: Cameron returned home with a bag of
                # candy to eat all night
                # long. What will Others want to do next?
                # A. great
                # B. buy the candy to eat
                # C. bored
                # Answer:
                hypotheses = options_sym
                premise = (
                    f"{args.multiple_choice_prompt} "
                    + f"Question: {premise}\nA. {options_text[0]}\nB. "
                    + f"{options_text[1]}\nC. {options_text[2]}\nAnswer:"
                )
            else:
                hypotheses = options_text
                premise = premise + uncond_premise

            examples_siqa += [
                {
                    "label": label,
                    "premise": premise,
                    "uncond_premise": uncond_premise,
                    "hypothesis0": hypotheses[0],
                    "hypothesis1": hypotheses[1],
                    "hypothesis2": hypotheses[2],
                }
            ]
    return examples_siqa


def winogrande_loader(path, args):
    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"
    examples_winogrande = []
    qa_path, label_path = path[0], path[1]

    with open(qa_path) as lines, open(label_path) as labels:
        for line, label_sym in zip(lines, labels):
            label = int(label_sym[0]) - 1
            line = json.loads(line)
            premise = line["sentence"]

            options_text = [line["option1"], line["option2"]]
            options_sym = ["A", "B"]

            if getattr(args, "multiple_choice_prompt", None) is not None:
                # Question: So _ plays video games because
                # Leslie has a lot of free time
                # while Nelson has to work all the time.
                # A. Leslie
                # B. Nelson
                # Answer:
                hypotheses = options_sym
                premise = (
                    f"{args.multiple_choice_prompt} "
                    + f"Question: {premise}\nA. {options_text[0]}\nB. "
                    + f"{options_text[1]}\nAnswer:"
                )
            else:
                hypotheses = options_text
                premise = premise + uncond_premise

            examples_winogrande += [
                {
                    "label": label,
                    "premise": premise,
                    "uncond_premise": uncond_premise,
                    "hypothesis0": hypotheses[0],
                    "hypothesis1": hypotheses[1],
                }
            ]
    return examples_winogrande


def date_understanding_loader(path, args):
    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"
    examples = []

    for one_path in path:
        with open(one_path) as json_file:
            data = json.load(json_file)
            task_prefix = data.get("task_prefix", "")
            for instance in data["examples"]:
                options_text = list(instance["target_scores"].keys())
                num_options = len(options_text)
                if (
                    args.num_options is not None
                    and num_options != args.num_options
                ):
                    continue
                options_sym = [chr(ord("A") + i) for i in range(num_options)]
                for target, score in instance["target_scores"].items():
                    if score == 1:
                        raw_label = target  # e.g., stare wars
                label = options_text.index(raw_label)
                premise = instance["input"]
                premise = task_prefix + premise

                if getattr(args, "multiple_choice_prompt", None) is not None:
                    # Question: "Which of the following is a
                    # humorous edit of this artist or movie name: 'star wars'?"
                    # A. stare wars
                    # B. stariwars
                    # C. ...
                    # D. ...
                    # Answer:
                    hypotheses = options_sym
                    # premise = f"{args.multiple_choice_prompt} Question: \
                    #     {premise}\nA. {options_text[0]}\nB. \
                    #         {options_text[1]}\nC. \
                    #             {options_text[2]}\nD. \
                    #                 {options_text[3]}\nE. \
                    #                     {options_text[4]}\nAnswer:"
                    premise = (
                        f"{args.multiple_choice_prompt} Question: {premise}\n"
                    )
                    for idx in range(num_options):
                        premise += f"{options_sym[idx]}. {options_text[idx]}\n"
                    premise += "Answer:"
                else:
                    hypotheses = options_text
                    premise = premise + uncond_premise
                example = [
                    {
                        "label": label,
                        "premise": premise,
                        "uncond_premise": uncond_premise,
                    }
                ]
                for idx in range(num_options):
                    example[0][f"hypothesis{idx}"] = hypotheses[idx]

                examples += example
    return examples


def anli_loader(path, args):
    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"
    examples = []
    options_text = ["entailment", "neutral", "contradiction"]
    options_sym = ["A", "B", "C"]
    num_options = 3

    for one_path in path:
        with open(one_path) as lines:
            for line in lines:
                line = json.loads(line)
                label = ["e", "n", "c"].index(line["label"])
                premise = f"{line['context']} {line['hypothesis']}"

                if getattr(args, "multiple_choice_prompt", None) is not None:
                    # Question: "Which of the following is a humorous \
                    # edit of this artist or movie name: 'star wars'?"
                    # A. entailment
                    # B. neutral
                    # C. contradiction
                    # Answer:
                    hypotheses = options_sym
                    # premise = f"{args.multiple_choice_prompt} Question: \
                    #     {premise}\nA. {options_text[0]}\nB. \
                    #         {options_text[1]}\nC. \
                    #             {options_text[2]}\nD. \
                    #                 {options_text[3]}\nE. \
                    #                     {options_text[4]}\nAnswer:"
                    premise = (
                        f"{args.multiple_choice_prompt} Question: {premise}\n"
                    )
                    for idx in range(num_options):
                        premise += f"{options_sym[idx]}. {options_text[idx]}\n"
                    premise += "Answer:"
                else:
                    hypotheses = options_text
                    premise = premise + uncond_premise
                example = [
                    {
                        "label": label,
                        "premise": premise,
                        "uncond_premise": uncond_premise,
                    }
                ]
                for idx in range(num_options):
                    example[0][f"hypothesis{idx}"] = hypotheses[idx]
                examples += example

    return examples


def generate_n_shot_demonstrations(n_shot_dataset):
    n_shot_demonstrations = ""
    for raw_instance in n_shot_dataset:
        presmise = raw_instance["premise"]
        answer_index = raw_instance["label"].item()
        answer = raw_instance[f"hypothesis{answer_index}"]
        n_shot_instance = f"{presmise}{answer}\n\n"
        n_shot_demonstrations += n_shot_instance
    return n_shot_demonstrations


def create_n_shot_splits(raw_dataset, n_shot_dataset, args):
    n_shot_demonstrations = ""
    if args.n_shot > 0:
        # few-shot setting: sample from train split, dev split (COPA),
        # or the only split (BB)
        if (
            n_shot_dataset is raw_dataset
        ):  # BB tasks: sampling from the only split, and use the rest
            raw_dataset = raw_dataset.train_test_split(
                test_size=args.n_shot, seed=args.seed
            )
            raw_dataset, n_shot_dataset = (
                raw_dataset["train"],
                raw_dataset["test"],
            )
        else:
            n_shot_dataset = n_shot_dataset.shuffle(seed=args.seed).select(
                range(args.n_shot)
            )
        n_shot_demonstrations = generate_n_shot_demonstrations(n_shot_dataset)

    if args.sample is not None and args.sample <= len(raw_dataset):
        # sample "sample" amount of data from raw_data
        raw_dataset = raw_dataset.shuffle(seed=args.seed).select(
            range(args.sample)
        )

    if args.n_shot > 0:
        # append n_shot_demonstrations to each input.
        raw_dataset = raw_dataset.map(
            lambda x: {"premise": n_shot_demonstrations + x["premise"]}
        )
    else:  # zero-shot: no need to return n_shot_dataset
        n_shot_dataset = None
    return raw_dataset, n_shot_dataset, n_shot_demonstrations


def generate_n_shot_poe_demonstrations(n_shot_dataset, num_of_options):
    # designed for multiple_choice_prompt
    alphabets = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"[:num_of_options]
    last_option = alphabets[-1]
    null_string = "[MASK]"
    n_shot_demonstrations = ""
    n_shot_poe_demonstrations = ""
    for raw_instance in n_shot_dataset:
        premise = raw_instance["premise"]
        answer_index = raw_instance["label"].item()
        answer = raw_instance[f"hypothesis{answer_index}"]
        n_shot_instance = f"{premise}{answer}\n\n"
        n_shot_demonstrations += n_shot_instance

        # for mcp: randomly mask options to [MASK]
        poe_premise = premise
        new_alphabets = alphabets.replace(alphabets[answer_index], "")
        num_of_mask_options = random.randint(1, num_of_options - 1)
        mask_option_symbols = random.sample(
            new_alphabets, num_of_mask_options
        )  # e.g., [B, C]
        for symbol in mask_option_symbols:
            option_start_index = poe_premise.rfind(f"{symbol}. ")
            if symbol == last_option:
                option_end_index = poe_premise.rfind("Answer:")
            else:
                option_end_index = poe_premise.rfind(
                    f"{alphabets[alphabets.index(symbol) + 1]}. "
                )
            option = poe_premise[option_start_index:option_end_index]
            poe_premise = poe_premise.replace(
                option, f"{symbol}. {null_string}\n"
            )

        n_shot_poe_instance = f"{poe_premise}{answer}\n\n"
        n_shot_poe_demonstrations += n_shot_poe_instance
    return n_shot_demonstrations, n_shot_poe_demonstrations


def vqa_loader(path, args):
    version_type = ""
    task_type = "MultipleChoice"
    data_type = "mscoco"
    data_subtype = "%s2014" % args.split
    ann_file = "%s/Annotations/%s%s_%s_annotations.json" % (
        path,
        version_type,
        data_type,
        data_subtype,
    )
    question_file = "%s/Questions/%s%s_%s_%s_questions.json" % (
        path,
        version_type,
        task_type,
        data_type,
        data_subtype,
    )
    img_dir = "%s/Images/%s/%s" % (path, data_type, data_subtype)
    alphabets = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    examples = []

    print("Loading annotations and questions...")
    anno = json.load(open(ann_file, "r"))
    ques = json.load(open(question_file, "r"))

    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"

    for i in range(len(anno["annotations"])):
        ans = anno["annotations"][i]["multiple_choice_answer"]
        img_id = anno["annotations"][i]["image_id"]
        # question_id = train_anno['annotations'][i]['question_id']
        image_path = os.path.join(
            img_dir, "COCO_%s2014_" % args.split + "%012d.jpg" % img_id
        )

        question = ques["questions"][i]["question"]
        mc_ans = ques["questions"][i]["multiple_choices"]
        label = mc_ans.index(ans)

        if getattr(args, "multiple_choice_prompt", None) is not None:
            hypotheses = mc_ans
            # Question: How does a bishop move from one place to another?
            # A. chess game
            # B. church
            # C. in a car
            # D. queen
            # E. cathedral
            # Answer:
            options = "\n".join(
                [f"{alphabets[i]}. {ans}" for i, ans in enumerate(mc_ans)]
            )
            premise = (
                f"{args.multiple_choice_prompt} "
                + f"Question: {question}\n{options}\nAnswer:"
            )
        else:
            hypotheses = mc_ans
            premise = question + uncond_premise

        example = [
            {
                "premise": premise,
                "image_path": image_path,
                "uncond_premise": uncond_premise,
                "label": label,
            }
        ]

        for idx, ans in enumerate(hypotheses):
            example[0][f"hypothesis{idx}"] = ans
        examples += example

    return examples


def scienceqa_loader(path, args):
    ann_file = "%s/ScienceQA_DATA/problems.json" % (path)
    img_dir = "%s/ScienceQA_DATA/%s" % (path, args.split)
    alphabets = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    examples = []

    print("Loading annotations and images...")
    anno = json.load(open(ann_file, "r"))
    ids = os.listdir(img_dir)
    anno = {id: anno[id] for id in ids}

    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"

    for i, (id, value) in enumerate(anno.items()):
        img_id = id
        question = value["question"]
        mc_ans = value["choices"]
        label = int(value["answer"])
        image_file = value["image"]

        if (not len(mc_ans) == args.num_options) or (image_file is None):
            continue

        image_path = os.path.join(os.path.join(img_dir, img_id), image_file)
        if getattr(args, "multiple_choice_prompt", None) is not None:
            hypotheses = mc_ans
            # Question: How does a bishop move from one place to another?
            # A. chess game
            # B. church
            # C. in a car
            # D. queen
            # Answer:
            options = "\n".join(
                [f"{alphabets[i]}. {ans}" for i, ans in enumerate(mc_ans)]
            )
            premise = (
                f"{args.multiple_choice_prompt} "
                + f"Question: {question}\n{options}\nAnswer:"
            )
        else:
            hypotheses = mc_ans
            premise = question + uncond_premise

        example = [
            {
                "premise": premise,
                "image_path": image_path,
                "uncond_premise": uncond_premise,
                "label": label,
            }
        ]

        for idx, ans in enumerate(hypotheses):
            example[0][f"hypothesis{idx}"] = ans
        examples += example
    print("Dataset Length: ", len(examples))
    return examples


def ai2d_loader(path, args):
    question_dir = "%s/ai2d/questions" % (path)
    imgDir = "%s/ai2d/images" % (path)
    alphabets = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    examples = []

    print("Loading annotations and images...")
    train_files = os.listdir(question_dir)

    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"

    for i, file in enumerate(train_files):
        anno = json.load(open(os.path.join(question_dir, file), "r"))
        questions = anno["questions"]
        imageName = anno["imageName"]
        for question, value in questions.items():
            mc_ans = value["answerTexts"]
            label = int(value["correctAnswer"])
            abcLabel = value["abcLabel"]

            if (
                (not len(mc_ans) == args.num_options)
                or (imageName is None)
                or abcLabel is True
            ):
                continue

            image_path = os.path.join(imgDir, imageName)
            if getattr(args, "multiple_choice_prompt", None) is not None:
                hypotheses = mc_ans
                # Question: How does a bishop move from one place to another?
                # A. chess game
                # B. church
                # C. in a car
                # D. queen
                # Answer:
                options = "\n".join(
                    [f"{alphabets[i]}. {ans}" for i, ans in enumerate(mc_ans)]
                )
                premise = (
                    f"{args.multiple_choice_prompt} "
                    + f"Question: {question}\n{options}\nAnswer:"
                )
            else:
                hypotheses = mc_ans
                premise = question + uncond_premise

            example = [
                {
                    "premise": premise,
                    "image_path": image_path,
                    "uncond_premise": uncond_premise,
                    "label": label,
                }
            ]

            for idx, ans in enumerate(hypotheses):
                example[0][f"hypothesis{idx}"] = ans
            examples += example
    print("Dataset Length: ", len(examples))
    return examples


def single_inference_loader(path, args):
    alphabets = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    examples = []

    print("Loading single question and image...")
    question = args.question
    mc_ans = args.choices
    label = args.label
    image_path = path

    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"

    if getattr(args, "multiple_choice_prompt", None) is not None:
        hypotheses = mc_ans
        # Question: How does a bishop move from one place to another?
        # A. chess game
        # B. church
        # C. in a car
        # D. queen
        # Answer:
        options = "\n".join(
            [f"{alphabets[i]}. {ans}" for i, ans in enumerate(mc_ans)]
        )
        premise = (
            f"{args.multiple_choice_prompt} "
            + f"Question: {question}\n{options}\nAnswer:"
        )
    else:
        hypotheses = mc_ans
        premise = question + uncond_premise

    example = [
        {
            "premise": premise,
            "image_path": image_path,
            "uncond_premise": uncond_premise,
            "label": label,
        }
    ]

    for idx, ans in enumerate(hypotheses):
        example[0][f"hypothesis{idx}"] = ans
    examples += example
    print("Dataset Length: ", len(examples))
    return examples


# Custom Dataloader
def custom_loader(path, args):
    # Annotation/Question file format:
    # {
    #     "COCO_train2014_000000000025": {
    #         "question": "What is the capital of France?",
    #         "choices": ["Paris", "London", "Berlin", "Madrid"],
    #         "answer": 0,
    #         "image": "COCO_train2014_000000000025.jpg",
    #     }
    # }
    ann_file = "%s/questions.json" % (path)

    # Image directory/file format:
    # images/COCO_train2014_000000000025.jpg
    img_dir = "%s/images" % (path)

    alphabets = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    examples = []

    print("Loading annotations and images...")
    anno = json.load(open(ann_file, "r"))

    if args.calibration_prompt is not None:
        uncond_premise = args.calibration_prompt
    else:
        uncond_premise = " the answer is:"

    for i, (question_id, value) in enumerate(anno.items()):
        question = value["question"]
        mc_ans = value["choices"]
        label = int(value["answer"])
        image_file = value["image"]

        if (not len(mc_ans) == args.num_options) or (image_file is None):
            continue

        image_path = os.path.join(img_dir, image_file)
        if getattr(args, "multiple_choice_prompt", None) is not None:
            hypotheses = mc_ans
            # Question: What is the capital of France?
            # A. Paris
            # B. London
            # C. Berlin
            # D. Madrid
            # Answer:
            options = "\n".join(
                [f"{alphabets[i]}. {ans}" for i, ans in enumerate(mc_ans)]
            )
            premise = (
                f"{args.multiple_choice_prompt} "
                + f"Question: {question}\n{options}\nAnswer:"
            )
        else:
            hypotheses = mc_ans
            premise = question + uncond_premise

        example = [
            {
                "premise": premise,
                "image_path": image_path,
                "uncond_premise": uncond_premise,
                "label": label,
            }
        ]

        for idx, ans in enumerate(hypotheses):
            example[0][f"hypothesis{idx}"] = ans
        examples += example
    print("Dataset Length: ", len(examples))
    return examples
